"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""
import warnings

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.impute import SimpleImputer as skImputer
from ..utils.stat_utils import which_columns_are_binary
from causallib.estimation import Matching


# TODO: Entire module might be redundant, now that scikit-learn supports missing values
#       in its preprocessing: https://scikit-learn.org/stable/whats_new/v0.20.html#highlights
#       The only support now needed is:
#       1) Transforming from numpy-array to pandas DataFrame in a pipeline, before specifying a causal model.
#       2) Possible generic support for causallib's additional `a` parameter, along with `X` and `y`.


class StandardScaler(BaseEstimator, TransformerMixin):
    """
    Standardize continuous features by removing the mean and scaling to unit variance while allowing nans.

        X = (X - X.mean()) / X.std()
    """

    def __init__(self, with_mean=True, with_std=True, ignore_nans=True):
        """

        Args:
            with_mean (bool): Whether to center the data before scaling.
            with_std (bool): Whether to scale the data to unit variance.
            ignore_nans (bool): Whether to ignore NaNs during calculation.
        """
        self.with_mean = with_mean
        self.with_std = with_std
        self.ignore_nans = ignore_nans

    def fit(self, X, y=None):
        """
        Compute the mean and std to be used for later scaling.

        Args:
            X (pd.DataFrame): The data used to compute the mean and standard deviation used for later scaling along the
                            features axis (axis=0).
            y: Passthrough for ``Pipeline`` compatibility.

        Returns:
            StandardScaler: A fitted standard-scaler
        """
        continuous_features = self._get_relevant_features(X)
        self._feature_mask_ = continuous_features

        if self.with_mean:
            means = X.loc[:, self._feature_mask_].mean(skipna=self.ignore_nans)
        else:
            means = pd.Series(0, index=continuous_features)
        self.mean_ = means

        if self.with_std:
            scales = X.loc[:, self._feature_mask_].std(skipna=self.ignore_nans)
        else:
            scales = pd.Series(1, index=continuous_features)
        self.scale_ = scales

        return self

    def transform(self, X, y='deprecated'):
        """
        Perform standardization by centering and scaling

        Args:
            X (pd.DataFrame): array-like, shape [n_samples, n_features] The data used to compute the mean and standard
                            deviation used for later scaling along the features axis (axis=0).
            y: Passthrough for ``Pipeline`` compatibility.X:

        Returns:
            pd.DataFrame: Scaled dataset.
        """
        # Taken from the sklearn implementation. Will probably need adjustment when a new scikit-learn version is out:
        if not isinstance(y, str) or y != 'deprecated':
            warnings.warn("The parameter y on transform() is deprecated since 0.19 and will be removed in 0.21",
                          DeprecationWarning)

        X = X.copy()  # type: pd.DataFrame
        if self.with_mean:
            X.loc[:, self._feature_mask_] -= self.mean_
        if self.with_std:
            X.loc[:, self._feature_mask_] /= self.scale_
        return X

    def inverse_transform(self, X):
        """
        Scale back the data to the original representation

        Args:
            X (pd.DataFrame): array-like, shape [n_samples, n_features] The data used to compute the mean and standard
                              deviation used for later scaling along the features axis (axis=0).

        Returns:
            pd.DataFrame: Un-scaled dataset.
        """
        X = X.copy()  # type: pd.DataFrame
        if self.with_std:
            X.loc[:, self._feature_mask_] *= self.scale_
        if self.with_mean:
            X.loc[:, self._feature_mask_] += self.mean_
        return X

    @staticmethod
    def _get_relevant_features(X):
        """
        Returns a binary mask specifying the continuous features to operate on.

        Args:
            X (pd.DataFrame): array-like, shape [n_samples, n_features] The data used to compute the mean and standard
                              deviation used for later scaling along the features axis (axis=0).

        Returns:
            pd.Index: a pd.Index with name of columns specifying which features to apply the transformation on.
        """
        # FIXME utilize sklearn.utils.multiclass.type_of_target()
        continuous_cols = X.columns[~which_columns_are_binary(X)]
        return continuous_cols


class MinMaxScaler(BaseEstimator, TransformerMixin):
    """
    Scales features to 0-1, allowing for NaNs.

        X_std = (X - X.min(axis=0)) / (X.max(axis=0) - X.min(axis=0))
    """

    def __init__(self, only_binary_features=True, ignore_nans=True):
        """

        Args:
            only_binary_features (bool): Whether to apply only on binary features or across all.
            ignore_nans (bool): Whether to ignore NaNs during calculation.
        """
        self.only_binary_features = only_binary_features
        self.ignore_nans = ignore_nans

    def fit(self, X, y=None):
        """
        Compute the minimum and maximum to be used for later scaling.

        Args:
            X (pd.DataFrame): array-like, shape [n_samples, n_features] The data used to compute the mean and standard
                              deviation used for later scaling along the features axis (axis=0).
            y: Passthrough for ``Pipeline`` compatibility.

        Returns:
            MinMaxScaler: a fitted MinMaxScaler
        """
        feature_mask = self._get_relevant_features(X)
        self._feature_mask_ = feature_mask

        self.min_ = X.min(skipna=self.ignore_nans)[feature_mask]
        self.max_ = X.max(skipna=self.ignore_nans)[feature_mask]
        self.scale_ = self.max_ - self.min_

        # if feature_mask.size != X.shape[1]:
        #     self.scale_[~feature_mask] = 1
        #     self.min_[~feature_mask] = 0
        #     self.max_[~feature_mask] = 1

        return self

    def inverse_transform(self, X):
        """
        Scaling chosen features of X to the range of 0 - 1.

        Args:
            X (pd.DataFrame): array-like, shape [n_samples, n_features] Input data that will be transformed.

        Returns:
            pd.DataFrame: array-like, shape [n_samples, n_features]. Transformed data.
        """
        # No warning for y, since there's no y variable.
        # This correpsonds to function signature in scikit-learn's code base
        X = X.copy()  # type: pd.DataFrame
        X.loc[:, self._feature_mask_] *= self.scale_
        X.loc[:, self._feature_mask_] += self.min_
        return X

    def transform(self, X):
        """
        Undo the scaling of X according to feature_range.

        Args:
            X (pd.DataFrame): array-like, shape [n_samples, n_features] Input data that will be transformed.

        Returns:
            pd.DataFrame: array-like, shape [n_samples, n_features]. Transformed data.
        """

        X = X.copy()  # type: pd.DataFrame
        X.loc[:, self._feature_mask_] -= self.min_
        X.loc[:, self._feature_mask_] /= self.scale_
        return X

    def _get_relevant_features(self, X):
        """
        Returns a binary mask specifying the features to operate on (either all features or binary features if
        self.only_binary_features is True.

        Args:
            X (pd.DataFrame): array-like, shape [n_samples, n_features] The data used to compute the mean and standard
                            deviation used for later scaling along the features axis (axis=0).

        Returns:
            pd.Index: a binary mask specifying which features to apply the transformation on.
        """
        if self.only_binary_features:
            feature_mask = which_columns_are_binary(X)
        else:
            feature_mask = np.ones(X.shape[1], dtype=bool)
        return feature_mask


class Imputer(skImputer):
    def transform(self, X):
        X_transformed = super().transform(X.values)
        X_transformed = pd.DataFrame(
            X_transformed, index=X.index, columns=X.columns)
        return X_transformed


class PropensityTransformer(BaseEstimator, TransformerMixin):

    def __init__(self, learner, include_covariates=False):
        """Transform covariates by adding/replacing with the propensity score.
    
        Args:
            learner (sklearn.estimator) : A learner implementing `fit` and 
                `predict_proba` to use for predicting the propensity score.
            include_covariates (bool) : Whether to return the original
                covariates alongside the "propensity" column.

        """
        self.include_covariates = include_covariates
        self.learner = learner

    def fit(self, X, a):
        self.learner.fit(X, a)
        return self

    def transform(self, X, treatment_values=None):
        """Append propensity or replace covariates with propensity.

        Args:
            X (pd.DataFrame): A DataFrame of samples to transform. This will be
                input to the learner trained by fit. If the columns are 
                different, the results will not be valid.
            treatment_values (Any | None): A desired value/s to extract
                propensity to (i.e. probabilities to what treatment value
                should be calculated). If not specified, then the maximal
                treatment value is chosen. This is since the usual case is of
                treatment (A=1) control (A=0) setting.
            
        Returns:
            pd.DataFrame : DataFrame with a "propensity" column. 
            If "include_covariates" is `True`, it will include all of the 
            original features plus "propensity", else it will only have the 
            "propensity" column.

        """
        treatment_values = 1 if treatment_values is None else treatment_values

        res = self.learner.predict_proba(X)[:, treatment_values]
        res = pd.DataFrame(res, index=X.index, columns=["propensity"])
        if self.include_covariates:
            res = X.join(res)
        return res


class MatchingTransformer(object):

    def __init__(
        self,
        propensity_transform=None,
        caliper=None,
        with_replacement=True,
        n_neighbors=1,
        matching_mode="both",
        metric="mahalanobis",
        knn_backend="sklearn",
    ):
        """Transform data by removing poorly matched samples.

        Args:
            propensity_transform (causallib.transformers.PropensityTransformer):
                an object for data preprocessing which adds the propensity
                score as a feature (default: None)
            caliper (float) : maximal distance for a match to be accepted. If
                not defined, all matches will be accepted. If defined, some
                samples may not be matched and their outcomes will not be
                estimated. (default: None)
            with_replacement (bool): whether samples can be used multiple times
                for matching. If set to False, the matching process will optimize
                the linear sum of distances between pairs of treatment and
                control samples and only `min(N_treatment, N_control)` samples
                will be estimated. Matching with no replacement does not make
                use of the `fit` data and is therefore not implemented for
                out-of-sample data (default: True)
            n_neighbors (int) : number of nearest neighbors to include in match.
                Must be 1 if `with_replacement` is `False.` If larger than 1, the
                estimate is calculated using the `regress_agg_function` or 
                `classify_agg_function` across the `n_neighbors`. Note that when
                the `caliper` variable is set, some samples will have fewer than
                `n_neighbors` matches. (default: 1).
            matching_mode (str) : Direction of matching: `treatment_to_control`,
                `control_to_treatment` or `both` to indicate which set should
                be matched to which. All sets are cross-matched in `match`
                and when `with_replacement` is `False` all matching modes 
                coincide. With replacement there is a difference.
            metric (str) : Distance metric string for calculating distance
                between samples. Note: if an external built `knn_backend`
                object with a different metric is supplied, `metric` needs to
                be changed to reflect that, because `Matching` will set its 
                inverse covariance matrix if "mahalanobis" is set. (default: 
                "mahalanobis", also supported: "euclidean")
            knn_backend (str or callable) : Backend to use for nearest neighbor
                search. Options are "sklearn"  or a callable  which returns an 
                object implementing `fit`, `kneighbors` and `set_params` 
                like the sklearn `NearestNeighbors` object. (default: "sklearn"). 

        """
        self.matching = Matching(
            propensity_transform=propensity_transform,
            caliper=caliper,
            with_replacement=with_replacement,
            n_neighbors=n_neighbors,
            matching_mode=matching_mode,
            metric=metric,
            knn_backend=knn_backend,
        )

    def fit(self, X, a, y):
        """Fit data to transform

        This function loads the data for matching and must be called before
        `transform`. For convenience, consider using `fit_transform`.

        Args:
            X (pd.DataFrame): DataFrame of shape (n,m) containing m covariates
                for n samples.
            a (pd.Series): Series of shape (n,) containing discrete treatment
                values for the n samples.
            y (pd.Series): Series of shape (n,) containing outcomes for
                the n samples.

        Returns:
            self (MatchingTransformer) : Fitted object
        """
        self.matching.fit(X, a, y)

        return self

    def transform(self, X, a, y):
        """Transform data by restricting it to samples which are matched

        Following a matching process, not all of the samples will find matches.
        Transforming the data by only allowing samples in treatment that have
        close matches in control, or in control that have close matches in
        treatment can make other causal methods more effective. This function 
        will call `match` on the underlying Matching object.

        The attribute `matching_mode` changes the behavior of this function.
        If set to `control_to_treatment` each control will attempt to find a
        match among the treated, hence the transformed data will have a maximum
        size of N_c + min(N_c,N_t).
        If set to `treatment_to_control`, each treatment will attempt to find a
        match among the control and the transformed data will have a maximum
        size of N_t + min(N_c,N_t).
        If set to `both`, both matching operations will be executed and if a
        sample succeeds in either direction it will be included, hence the
        maximum size of the transformed data will be `len(X)`.

        If `with_replacement` is `False`, `matching_mode` does not change the
        behavior. There will be up to `min(N_c,N_t)` samples in
        the returned DataFrame, regardless.

        Args:
            X (pd.DataFrame): DataFrame of shape (n,m) containing m covariates
                for n samples.
            a (pd.Series): Series of shape (n,) containing discrete treatment
                values for the n samples.
            y (pd.Series): Series of shape (n,) containing outcomes for
                the n samples.

        Raises:
            NotImplementedError: Raised if a value of attribute `matching_mode`
            other than the supported values is set.

        Returns:
            Xm (pd.DataFrame): Covariates of samples that were matched
            am (pd.Series): Treatment values of samples that were matched
            ym (pd.Series): Outcome values of samples that were matched

        """
        self.matching.match(X, a, use_cached_result=True)
        matched_sample_indices = self.find_indices_of_matched_samples(X, a)
        X = X.loc[matched_sample_indices]
        a = a.loc[matched_sample_indices]
        y = y.loc[matched_sample_indices]
        return X, a, y

    def find_indices_of_matched_samples(self, X, a):
        """Find indices of samples which matched successfully.

        Given a DataFrame of samples `X` and treatment assignments `a`, return
        a list of indices of samples which matched successfully.

        Args:
            X (pd.DataFrame): Covariates of samples
            a (pd.Series): Treatment assignments

        Returns:
            pd.Series: indices of matched samples to be passed to `X.loc` 
        """

        matching_weights = self.matching.matches_to_weights()
        matches_mask = self._filter_matching_weights_by_mode(matching_weights)
        return matches_mask

    def _filter_matching_weights_by_mode(self, matching_weights):
        if self.matching.matching_mode == "control_to_treatment":
            matches_mask = matching_weights.control_to_treatment
        elif self.matching.matching_mode == "treatment_to_control":
            matches_mask = matching_weights.treatment_to_control
        elif self.matching.matching_mode == "both":
            matches_mask = matching_weights.sum(axis=1)
        else:
            raise NotImplementedError("Matching mode {} not supported".format(
                    self.matching.matching_mode))
        matches_mask = matches_mask.astype(bool)
        return matches_mask

    def fit_transform(self, X, a, y):
        """Match data and return matched subset.

        This is a convenience method, calling `fit` and `transform` at once.
        For details, see documentation of each function.

        Args:
            X (pd.DataFrame): DataFrame of shape (n,m) containing m covariates
                for n samples.
            a (pd.Series): Series of shape (n,) containing discrete treatment
                values for the n samples.
            y (pd.Series): Series of shape (n,) containing outcomes for
                the n samples.

        Returns:
            Xm (pd.DataFrame): Covariates of samples that were matched
            am (pd.Series): Treatment values of samples that were matched
            ym (pd.Series): Outcome values of samples that were matched
        """
        self.fit(X, a, y)
        return self.transform(X, a, y)

    def set_params(self, **kwargs):
        """Set parameters of matching engine. Supported parameters are:

        Keyword Args:
            propensity_transform (causallib.transformers.PropensityTransformer):
                an object for data preprocessing which adds the propensity
                score as a feature (default: None)
            caliper (float) : maximal distance for a match to be accepted
                (default: None)
            with_replacement (bool): whether samples can be used multiple times
                for matching (default: True)
            n_neighbors (int) : number of nearest neighbors to include in match.
                Must be 1 if `with_replacement` is False (default: 1).
            matching_mode (str) : Direction of matching: `treatment_to_control`,
                `control_to_treatment` or `both` to indicate which set should
                be matched to which. All sets are cross-matched in `match`
                and without replacement there is no difference in outcome,
                but with replacement there is a difference and it impacts
                the results of `transform`.
            metric (str) : Distance metric string for calculating
                distance between samples (default: "mahalanobis",
                    also supported: "euclidean")
            knn_backend (str or callable) : Backend to use for nearest neighbor
                search. Options are "sklearn"  or a callable  which returns an 
                object implementing `fit`, `kneighbors` and `set_params` like
                the sklearn `NearestNeighbors` object. (default: "sklearn"). 


        Returns:
            self: (MatchingTransformer) object with new parameters set

        """
        supported_params = [
            "propensity_transform",
            "caliper",
            "n_neighbors",
            "metric",
            "with_replacement",
            "matching_mode",
            "knn_backend",
        ]
        for key, value in kwargs.items():
            if key in supported_params:
                self.matching.__setattr__(key, value)
            else:
                warnings.warn(
                    "Received unsupported parameter: {}. Nothing done.".format(key))
        return self



# (C) Copyright 2021 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import faiss


class FaissNearestNeighbors:

    def __init__(self,
                 metric="mahalanobis",
                 index_type="flatl2", n_cells=100, n_probes=10):
        """NearestNeighbors object utilizing the faiss library for speed

        Implements the same API as sklearn but runs 5-10x faster. Utilizes the 
        `faiss` library https://github.com/facebookresearch/faiss . Tested with 
        version 1.7.0. If `faiss-gpu` is installed from pypi, GPU acceleration
        will be used if available. 

        Args:
            metric (str) :  Distance metric for finding nearest neighbors
                (default: "mahalanobis")
            index_type (str) : Index type within faiss to use
                (supported: "flatl2" and "ivfflat")
            n_cells (int) : Number of voronoi cells (only used for "ivfflat",
                default: 100)
            n_probes (int) : Number of voronoi cells to search in
                (only used for "ivfflat", default: 10)
        Attributes (after running `fit`):
            index_ : the faiss index fit from the data. For details about
            faiss indices, see the faiss documentation at 
            https://github.com/facebookresearch/faiss/wiki/Faiss-indexes .
        """
        self.metric = metric
        self.n_cells = n_cells
        self.n_probes = n_probes
        self.index_type = index_type

    def fit(self, X):
        """Create faiss index and train with data.

        Args:
            X (np.array): Array of N samples of shape (NxM)

        Returns:
            self: Fitted object
        """
        X = self._transform_covariates(X)
        if self.index_type == "flatl2":
            self.index_ = faiss.IndexFlatL2(X.shape[1])
            self.index_.add(X)
        elif self.index_type == "ivfflat":
            quantizer = faiss.IndexFlatL2(X.shape[1])
            n_cells = max(1, min(self.n_cells, X.shape[0]//200))
            n_probes = min(self.n_probes, n_cells)
            self.index_ = faiss.IndexIVFFlat(
                quantizer, X.shape[1], n_cells)
            self.index_.train(X)
            self.index_.nprobe = n_probes
            self.index_.add(X)
        else:
            raise NotImplementedError(
                "Index type {} not implemented. Please select"
                "one of [\"flatl2\", \"ivfflat\"]".format(self.index_type))
        return self

    def kneighbors(self, X, n_neighbors=1):
        """Find the k nearest neighbors of each sample in X

        Args:
            X (np.array):  Array of shape (N,M) of samples to search
                for neighbors of. M must be the same as the fit data.
            n_neighbors (int, optional): Number of neighbors to find.
                Defaults to 1.

        Returns:
            (distances, indices): Two np.array objects of shape (N,n_neighbors)
                containing the distances and indices of the closest neighbors.
        """
        X = self._transform_covariates(X)
        distances, indices = self.index_.search(X, n_neighbors)
        # faiss returns euclidean distance squared
        return np.sqrt(distances), indices

    def _transform_covariates(self, X):
        if self.metric == "mahalanobis":
            if not hasattr(self, "VI"):
                raise AttributeError("Set inverse covariance VI first.")
            X = np.dot(X, self.VI.T)
        return np.ascontiguousarray(X).astype("float32")

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            if parameter == "metric_params":
                self.set_params(**value)
            else:
                self._setattr(parameter, value)
        return self

    def get_params(self, deep=True):
        # `deep` plays no role because there are no sublearners
        params_to_return = ["metric", "n_cells", "n_probes", "index_type"]
        return {i: self.__getattribute__(i) for i in params_to_return}

    def _setattr(self, parameter, value):
        # based on faiss docs https://github.com/facebookresearch/faiss/wiki/MetricType-and-distances
        if parameter == "VI":
            value = np.linalg.inv(value)
            chol = np.linalg.cholesky(value)
            cholvi = np.linalg.inv(chol)
            value = cholvi
        setattr(self, parameter, value)



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Sep 9, 2021
"""

import warnings
from .base_estimator import IndividualOutcomeEstimator
import pandas as pd
from copy import deepcopy
from sklearn.dummy import DummyClassifier


class XLearner(IndividualOutcomeEstimator):
    """
    An X-learner model for causal inference (künzel et al. 2018. pnas, https://www.pnas.org/content/116/10/4156).
    Uses two outcome estimators. The first is used to calculate the response while the second is used invertly to
    calculate the treatment which is averaged according to the propensity of the treatment assignment.
    """

    def __init__(self, outcome_model, effect_model=None, treatment_model=None, predict_proba=True, effect_types='diff'):
        """
        Args:
            outcome_model (IndividualOutcomeEstimator): Initialized causallib estimator that will be used to predict the
                                                        outcome of each treatment  given a case and a certain. To adhere
                                                         to the XLearner algorithm a StratifiedStandardization object
                                                         should be used for both outcome and cate model initialized
                                                         with comparable sklearn learners. Xlearner algorithm is suitable
                                                         for a binary outcome, if a non binary outcome will be used the
                                                         class will view the last outcome versus the rest as the binary
                                                         outcome.
            effect_model (IndividualOutcomeEstimator | None): Initialized causallib estimator that will be used to predict
                                                            the treatment effect of each case. The treatment effect
                                                            is estimated on the observed set using the outcome model
                                                            if the treatment effect is continuous use a regression model.
                                                            The default estimator is cloned from the outcome model.
                                                            The cloning is done after the outcome model is fitted to
                                                            enable warm start of the cate model by the outcome model
                                                            if outcome_model has its warm_start attribute on.
            treatment_model: Initialized sklearn prediction model that will predict the probability of each treatment.
                             Xlearner algorithm is suitable for binary treatment.
            predict_proba (bool) : In case the outcome task is classification and in case `learner` supports the
                                  operation, if True - prediction will utilize learner's `predict_proba` or
                                  `decision_function` which returns a continuous matrix of size (n_samples, n_classes).
                                  If False - `predict` will be used and return value will be based on a vector of class
                                  classifications. Xlearner effect estimation (in the case of binary effect)
                                requires the outcome estimator to predict probabilities of classification
                                (predict_proba=True)
            effect_types (str): string from the set of EffectEstimator.CALCULATE_EFFECT keys
                        if none the sklearn DummyClassifier with prior strategy will be used.
        """
        self.outcome_model = outcome_model
        self.effect_model = effect_model
        if treatment_model is None:
            treatment_model = DummyClassifier(strategy="prior")
        self.treatment_model = treatment_model
        self.effect_types = effect_types
        learner = {'outcome_model': self.outcome_model,
                   'cate_model': self.effect_model,
                   'treatment_model': self.treatment_model}
        super(XLearner, self).__init__(learner, predict_proba=predict_proba)

    def estimate_effect(self, X, a, agg="population", predict_proba=None, effect_types=None):
        """Estimates the causal effect between treatment groups.

        Args:
            X (pd.DataFrame): Covariates to predict on.
            a (pd.Series): Corresponding treatment assignment to utilize for prediction.
                           Assumes treated group is coded as 1, and control group as 0.
            agg (str): Either "population" or "individual" - whether to calculate individual effect or population
                       effect.
            predict_proba (bool | None): In case the outcome task is classification and in case `learner` supports the
                                  operation, if True - prediction will utilize learner's `predict_proba` or
                                  `decision_function` which returns a continuous matrix of size (n_samples, n_classes).
                                  If False - `predict` will be used and return value will be based on a vector of class
                                  classifications. If None, will use the object's initialized predict_proba value
            effect_types (None): IGNORED

        Returns:
             pd.Series: the estimated causal effect
        """
        if effect_types is not None:
            warnings.warn("Effect type is determined during fit")
        weighted_cat = self._estimate_weighted_effect(X, a, predict_proba=predict_proba)
        effect = super(XLearner, self).estimate_effect(
            weighted_cat[1], (-1) * weighted_cat[0], agg=agg, effect_types='diff'
        )
        # The effect type here is used for the weighted sum according to the treatment probability
        # the effect type is determined during fit so we need to modify the outputted data frame
        # if agg is population we get a single values series and we can use rename if agg is
        # individual we get a data frame that we squeeze into a series and rename. If we will
        # squeeze the population series we will get a scalar not a series so we need the condition..
        if agg == "population":
            effect.rename({'diff': self.effect_types})
        else:
            effect = effect.squeeze().rename(self.effect_types)
        return effect

    def _estimate_weighted_effect(self, X, a, predict_proba=None):
        """

        Args:
            X (pd.DataFrame): Covariates to predict on.
            a (pd.Series): Corresponding treatment assignment to utilize for prediction.
            predict_proba (bool | None): In case the outcome task is classification and in case `learner` supports the
                                  operation, if True - prediction will utilize learner's `predict_proba` or
                                  `decision_function` which returns a continuous matrix of size (n_samples, n_classes).
                                  If False - `predict` will be used and return value will be based on a vector of class
                                  classifications. If None, will use the object's initialized predict_proba value

        Returns (pd.Series): The propensity weighted treatment effect
        """
        predict_proba = self.predict_proba if predict_proba is None else predict_proba
        outcomes = self.effect_model.estimate_individual_outcome(X, a, predict_proba=predict_proba)
        propensity = self.treatment_model.predict_proba(X)
        return outcomes * propensity[:, ::-1]

    def estimate_individual_outcome(self, X, a, treatment_values=None, predict_proba=None):
        warnings.warn("If you're trying to estimate the individual outcome effect from individual outcome "
                      "prediction directly.\n You should use stratified or standarized estimator \n "
                      "XLearner effect estimation is obtained using the estimate_effect method")

        return self.outcome_model.estimate_individual_outcome(
            X, a, treatment_values=treatment_values, predict_proba=predict_proba)

    def fit(self, X, a, y, sample_weight=None, predict_proba=None):
        """
        Trains a causal model from observed data.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y (pd.Series): Observed outcome of size (num_subjects,).
            sample_weight: To be passed to the underlining outcome model fit method.
            predict_proba (bool | None): In case the outcome task is classification and in case `learner` supports the
                                  operation, if True - prediction will utilize learner's `predict_proba` or
                                  `decision_function` which returns a continuous matrix of size (n_samples, n_classes).
                                  If False - `predict` will be used and return value will be based on a vector of class
                                  classifications. If None, will use the object's initialized predict_proba value

        Returns:
            IndividualOutcomeEstimator: A causal model with an inner models fitted.
        """

        self.outcome_model.fit(X, a, y,  sample_weight=sample_weight)
        imp_te = self._estimate_imputed_treatment_effect(X, a, y, predict_proba=predict_proba)
        # Copying the outcome model post fit enables the user to use warm start that may benefit the cate model
        if self.effect_model is None:
            self.effect_model = deepcopy(self.outcome_model)
        self.effect_model.fit(X, a, imp_te, sample_weight=sample_weight)
        self.treatment_model.fit(X, a, sample_weight=sample_weight)
        return self

    def _estimate_imputed_treatment_effect(self, X, a, y, predict_proba=None):
        """
        Calculates the imputed treatment effect

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y (pd.Series): Observed outcome of size (num_subjects,).
            predict_proba (bool | None): In case the outcome task is classification and in case `learner` supports the
                                  operation, if True - prediction will utilize learner's `predict_proba` or
                                  `decision_function` which returns a continuous matrix of size (n_samples, n_classes).
                                  If False - `predict` will be used and return value will be based on a vector of class
                                  classifications. If None, will use the object's initialized predict_proba value

        Returns:
             pd.Series: The imputed treatment effect for each observation
        """
        ind_outcomes = self._obtain_ind_outcomes(X, a, y, predict_proba=predict_proba)
        return self._obtain_imputed_treatment_effect(ind_outcomes, a, y)

    def _obtain_ind_outcomes(self, X, a, y, predict_proba=None):
        """
        This method returns the outcomes model individual outcomes as a two column matrix
        for classification sklearn models predict_proba=True returns a matrix of probability for each y value
        for complying with the XLearner algorithm we extract only the probabilities for the maximal value of y
        (which is usually y==1 in binary tasks).

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y (pd.Series): Observed outcome of size (num_subjects,).
            predict_proba (bool | None): In case the outcome task is classification and in case `learner` supports the
                                  operation, if True - prediction will utilize learner's `predict_proba` or
                                  `decision_function` which returns a continuous matrix of size (n_samples, n_classes).
                                  If False - `predict` will be used and return value will be based on a vector of class
                                  classifications. If None, will use the object's initialized predict_proba value

        Returns:
            pd.DataFrame: DataFrame which columns are treatment values and rows are individuals: each column is a vector
              size (num_samples,) that contains the estimated outcome for each individual under the
              treatment value in the corresponding key.
        """
        predict_proba = self.predict_proba if predict_proba is None else predict_proba
        ind_outcomes = self.outcome_model.estimate_individual_outcome(X, a, predict_proba=predict_proba)
        if isinstance(ind_outcomes.columns, pd.MultiIndex):
            # predict_proba is True, columns are treatment-values over outcome-values
            ind_outcomes = ind_outcomes.xs(y.max(), axis="columns", level=-1, drop_level=True)
        return ind_outcomes

    def _obtain_imputed_treatment_effect(self, ind_outcomes, a, y):
        """
        Imputes outcome predictions with observed values for the factual outcome
        and calculates the effect.

        Args:
            ind_outcomes (pd.DataFrame): DataFrame which columns are treatment values and rows are individuals: each
            column is a vector size (num_samples,) that contains the estimated outcome for each individual under the
            treatment value in the corresponding key.
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y (pd.Series): Observed outcome of size (num_subjects,).

        Returns:
             pd.Series: The imputed treatment effect for each observation calculated according to the treatment
        """
        ind_outcomes.loc[a == 0, 0] = y.loc[a == 0]
        ind_outcomes.loc[a == 1, 1] = y.loc[a == 1]
        effect_func = self.CALCULATE_EFFECT[self.effect_types]
        ind_effect = effect_func(ind_outcomes[1], ind_outcomes[0])
        return ind_effect



# (C) Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Created on Oct 30, 2019

from sklearn.base import clone
from sklearn.model_selection import KFold, cross_val_predict, ParameterGrid
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn.preprocessing import LabelEncoder
from sklearn.utils.class_weight import compute_class_weight

import numpy as np


def select_classifier(model, X, A, n_splits=5, loss_type='01', seed=None):
    """Utility for selecting best classifier using cross-validation.

    Args:
        model: Either one of: scikit-learn classifier, scikit-learn SearchCV model (GridSearchCV, RandomizedSearchCV),
                              list of classifiers.
        X (np.ndarray): Covariate matrix size (num_samples, num_features)
        A (np.ndarray): binary labels indicating the source and target populations (num_samples,)
        n_splits (int): number of splits in cross-validation. relevant only if list of classifiers is passed.
        loss_type (str): name of loss metric to select classifier by. Either '01' for zero-one loss, otherwise
                         cross-entropy is used (and classifiers must implement predict_proba).
                         relevant only if list of classifiers is passed.
        seed (int): random seed for cross-validation split. relevant only if list of classifiers is passed.

    Returns:
        classifier: best performing classifier on validation set.
    """
    if isinstance(model, (GridSearchCV, RandomizedSearchCV)):
        selected_model = _select_classifier_from_sk_search(model, X, A)
    elif isinstance(model, list):
        selected_model = _select_classifier_from_list(candidates=model, X=X, A=A, n_splits=n_splits, seed=seed,
                                                      loss_type=loss_type)
    elif isinstance(model, dict):
        selected_model = _select_classifier_from_grid(X=X, A=A, n_splits=n_splits, seed=seed, **model,
                                                      loss_type=loss_type)
    else:  # A regular classifier was passed
        selected_model = model
    return selected_model


def _select_classifier_from_sk_search(estimator, X, A):
    """Return best model from a scikit-learn Search-estimator model.

    Args:
        estimator (GridSearchCV | RandomizedSearchCV): An initialized sklearn SearchCV classifier.
        X (np.ndarray): Covariate matrix size (num_samples, num_features)
        A (np.ndarray): binary labels indicating the source and target populations (num_samples,)

    Returns:
        classifier: model.best_estimator_ - best-performing classifier.
                    See scikit-learn's GridSearchCV and RandomizedSearchCV documentation for details on their return
                    values.
    """
    estimator.fit(X, A)
    best_estimator = clone(estimator.best_estimator_)
    return best_estimator


def _select_classifier_from_grid(estimator, X, A, param_grid, n_splits=5, seed=1, loss_type='01'):
    candidates = []
    for params in ParameterGrid(param_grid):
        estimator2 = clone(estimator)
        for key, value in params.items():
            setattr(estimator2, key, value)
        candidates.append(estimator2)

    return _select_classifier_from_list(candidates, X, A, n_splits=n_splits, seed=seed, loss_type=loss_type)


def _select_classifier_from_list(candidates, X, A, n_splits=5, seed=None, loss_type='01'):
    accuracies = np.zeros(len(candidates))

    class_weight = compute_class_weight(class_weight='balanced', classes=np.unique(A), y=A)[LabelEncoder().fit_transform(A)]

    if n_splits >= 2:
        cv = KFold(n_splits=n_splits, shuffle=True, random_state=seed)
        for model_idx, m in enumerate(candidates):
            if loss_type == '01':
                pred = cross_val_predict(m, X=X, y=A, cv=cv, fit_params={'sample_weight': class_weight}).reshape(-1)
            else:
                ps = cross_val_predict(m, X=X, y=A, cv=cv, fit_params={'sample_weight': class_weight},
                                       method='predict_proba')
                pred = ps[:, 1]
    else:
        for model_idx, m in enumerate(candidates):
            m.fit(X, A, sample_weight=class_weight)
            if loss_type == '01':
                pred = m.predict(X=X)
            else:
                pred = m.predict_proba(X=X)[:, 1]

    if loss_type == '01':
        accuracies[model_idx] = np.sum(class_weight[pred == A]) / np.sum(class_weight)
    else:
        logl = np.zeros(A.shape)
        logl[A == -1] = np.log(1.0 - pred[A == -1])
        logl[A == 1] = np.log(pred[A == 1])
        accuracies[model_idx] = np.sum(class_weight * logl) / np.sum(class_weight)

    i_best = np.argmax(accuracies)
    # print('accuracies =', accuracies, "accuracies-sorted", sorted(accuracies))
    # print('Selected model {} {}'.format(i_best, candidates[i_best]))
    return candidates[i_best]



import abc
import warnings
from typing import Type

import pandas as pd
import numpy as np
import statsmodels.api as sm
from sklearn.preprocessing import MinMaxScaler, OneHotEncoder
from sklearn.utils.multiclass import type_of_target

from .doubly_robust import BaseDoublyRobust
from causallib.estimation.base_estimator import IndividualOutcomeEstimator
from causallib.estimation.base_weight import PropensityEstimator
from causallib.utils.stat_utils import robust_lookup
from causallib.utils.general_tools import get_iterable_treatment_values, check_learner_is_fitted


class TMLE(BaseDoublyRobust):
    def __init__(self, outcome_model, weight_model,
                 outcome_covariates=None, weight_covariates=None,
                 reduced=False, importance_sampling=False,
                 glm_fit_kwargs=None,
                 ):
        """Targeted Maximum Likelihood Estimation.
        A model that takes an outcome model that was optimized to predict E[Y|X,A],
        and "retargets" ("updates") it to estimate E[Y^A|X] using a "clever covariate"
        constructed from the inverse propensity weights.

        Steps:
         1. Fit an outcome model Y=Q(X,A).
         2. Fit a weight model A=g(X,A).
         3. Construct a clever covariate using g(X,A).
         4. Fit a logistic regression model Q* to predict Y
            using g(X,A) as features and Q(X,A) as offset.
         5. Predict counterfactual outcome for treatment value `a` Q*(X,a)
            by plugging in Q(X,a) as offset, g(X,a) as covariate.

        Implements 4 flavours of TMLE controlled by the `reduced` and `importance_sampling` parameters.
        `importance_sampling=True` moves the clever covariate from being a feature to being a sample
        weight in the targeted regression.
        'reduced=True' use a clever covariate vector of 1s and -1s, therefore only good for binary treatment.
        Otherwise, the clever covariate are the entire IPW matrix and can be used for multiple treatments.

        References:
            * TMLE:
              Van Der Laan MJ, Rubin D. Targeted maximum likelihood learning. 2006.
              https://doi.org/10.2202/1557-4679.1043
            * TMLE with a vector version of clever covariate:
              Schuler MS, Rose S. Targeted maximum likelihood estimation for causal inference in observational studies.
              2017.
              https://doi.org/10.1093/aje/kww165
            * TMLE with a matrix version of clever covariate:
              Gruber S, van der Laan M. tmle: An R package for targeted maximum likelihood estimation. 2012.
              https://doi.org/10.18637/jss.v051.i13
            * TMLE with weighted regression and matrix of clever covariate:
              Gruber S, van der Laan M, Kennedy C. tmle: Targeted Maximum Likelihood Estimation. Cran documentation.
              https://cran.r-project.org/web/packages/tmle/index.html
            * TMLE for continuous outcomes
              Gruber S, van der Laan MJ. A targeted maximum likelihood estimator of a causal effect
              on a bounded continuous outcome. 2010.
              https://doi.org/10.2202/1557-4679.1260

        Args:
            outcome_model (IndividualOutcomeEstimator): An initial prediction of the outcome
            weight_model (PropensityEstimator): An IPW model predicting the treatment.
            outcome_covariates (array): Covariates to use for outcome model.
                                        If None - all covariates passed will be used.
                                        Either list of column names or boolean mask.
            weight_covariates (array): Covariates to use for weight model.
                                       If None - all covariates passed will be used.
                                       Either list of column names or boolean mask.
            reduced (bool): If `True` uses a vector version of the clever covariate
                            (rather than a matrix of all treatment values).
                            If `True` enforces a binary treatment assignment.
            importance_sampling (bool): If `True` moves the clever covariate from being
                                        a feature to being a weight in the regression.
            glm_fit_kwargs (dict): Additional kwargs for statsmodels' `GLM.fit()`.
              Can be used for example for refining the optimizers.
              see: https://www.statsmodels.org/stable/generated/statsmodels.genmod.generalized_linear_model.GLM.fit.html
        """
        super().__init__(
            outcome_model=outcome_model, weight_model=weight_model,
            outcome_covariates=outcome_covariates, weight_covariates=weight_covariates,
        )
        self.reduced = reduced
        self.importance_sampling = importance_sampling
        self.glm_fit_kwargs = {} if glm_fit_kwargs is None else glm_fit_kwargs

    def fit(self, X, a, y, refit_weight_model=True, **kwargs):
        # Initial outcome model:
        X_outcome = self._extract_outcome_model_data(X)
        self.outcome_model.fit(X_outcome, a, y)
        y_pred = self._outcome_model_estimate_individual_outcome(X, a)
        y_pred = robust_lookup(y_pred, a)  # Predictions on the observed

        # IPW to prepare covariates to fluctuate the initial estimator:
        weight_model_is_not_fitted = not check_learner_is_fitted(self.weight_model.learner)
        X_treatment = self._extract_weight_model_data(X)
        if refit_weight_model or weight_model_is_not_fitted:
            self.weight_model.fit(X_treatment, a)
        self.clever_covariate_ = _clever_covariate_factory(
            self.reduced, self.importance_sampling
        )(self.weight_model)
        exog = self.clever_covariate_.clever_covariate_fit(X, a)
        sample_weights = self.clever_covariate_.sample_weights(X, a)

        # Update the initial estimator with the IPW:
        self._validate_predict_proba_for_classification(y)
        # The re-targeting of the estimation is done through logistic regression,
        # which requires the target to be bounded between 0 and 1.
        # We force this bounding in case target is continuous.
        # See Gruber and van der Laan 2010: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3126669/
        self.target_scaler_ = TargetMinMaxScaler(feature_range=(0, 1))
        self.target_scaler_.fit(y)
        y = self.target_scaler_.transform(y)
        y_pred = self.target_scaler_.transform(y_pred)
        y_pred = _logit(y_pred)  # Used as offset in logit-space
        # Statsmodels supports logistic regression with continuous (0-1 bounded) targets
        # so can be used with non-binary (but scaled) response variable (`y`)
        # targeted_outcome_model = sm.Logit(
        #     endog=y, exog=clever_covariate, offset=y_pred,
        # ).fit(method="lbfgs")
        # GLM supports weighted regression, while Logit doesn't.
        targeted_outcome_model = sm.GLM(
            endog=y, exog=exog, offset=y_pred, freq_weights=sample_weights,
            family=sm.families.Binomial(),
            # family=sm.families.Binomial(sm.genmod.families.links.logit)
        ).fit(**self.glm_fit_kwargs)
        self.targeted_outcome_model_ = targeted_outcome_model

        return self

    def estimate_individual_outcome(self, X, a, treatment_values=None, predict_proba=None):
        potential_outcomes = self._outcome_model_estimate_individual_outcome(X, a)

        res = {}
        for treatment_value in get_iterable_treatment_values(treatment_values, a):
            potential_outcome = potential_outcomes[treatment_value]
            potential_outcome = self.target_scaler_.transform(potential_outcome)
            potential_outcome = _logit(potential_outcome)
            treatment_assignment = self.clever_covariate_.clever_covariate_inference(X, a, treatment_value)
            counterfactual_prediction = self.targeted_outcome_model_.predict(
                treatment_assignment, offset=potential_outcome,
            )
            counterfactual_prediction = self.target_scaler_.inverse_transform(counterfactual_prediction)
            res[treatment_value] = counterfactual_prediction

        res = pd.concat(res, axis="columns", names=[a.name or "a"])
        return res

    def _outcome_model_estimate_individual_outcome(self, X, a):
        """Standardize output for continuous `outcome_model` with `predict` with
        binary `outcome_model` with `predict_proba`"""
        # Force `predict_proba` so if `outcome_model.learner` is a classifier,
        # It will produce continuous scores.
        # For zeros and ones, logit(y_pred) will break with `inf`
        potential_outcomes = self.outcome_model.estimate_individual_outcome(X, a, predict_proba=True)

        is_predict_proba_classification_result = isinstance(potential_outcomes.columns, pd.MultiIndex)
        if is_predict_proba_classification_result:
            # Classification `outcome_model` with `predict_proba=True` returns
            # a MultiIndex treatment-values (`a`) over outcome-values (`y`)
            # Extract the prediction for the maximal outcome class
            # (probably class `1` in binary classification):
            outcome_values = potential_outcomes.columns.get_level_values(level=-1)
            potential_outcomes = potential_outcomes.xs(
                outcome_values.max(), axis="columns", level=-1, drop_level=True,
            )
        # TODO: move scale + logit here? since they always go together
        return potential_outcomes

    def _validate_predict_proba_for_classification(self, y):
        if type_of_target(y) != "continuous" and not self.outcome_model.predict_proba:
            warnings.warn(
                "`predict_proba` should be used in `outcome_model` if outcome type "
                "is not continuous. TMLE will force the use of `predict_proba`.",
                UserWarning
            )


class BaseCleverCovariate:
    def __init__(self, weight_model):
        self.weight_model = weight_model

    @abc.abstractmethod
    def clever_covariate_fit(self, X, a):
        raise NotImplementedError

    @abc.abstractmethod
    def clever_covariate_inference(self, X, a, treatment_value):
        raise NotImplementedError

    @abc.abstractmethod
    def sample_weights(self, X, a):
        raise NotImplementedError


class CleverCovariateFeatureVector(BaseCleverCovariate):
    """Clever covariate uses a signed vector of inverse propensity weights,
    with control group have their weights negated.
    The vector is then used as a predictor to the targeting regression.

    References:
        * Schuler MS, Rose S. Targeted maximum likelihood estimation for causal inference in observational studies. 2017
          https://doi.org/10.1093/aje/kww165
    """
    def clever_covariate_fit(self, X, a):
        if a.nunique() != 2:
            raise AssertionError("Can only apply model on a binary treatment")
        w = self.weight_model.compute_weights(X, a)
        w[a == 0] *= -1
        return w

    def clever_covariate_inference(self, X, a, treatment_value):
        weight_matrix = self.weight_model.compute_weight_matrix(X, a)
        w = weight_matrix[treatment_value]
        if treatment_value == 0:
            w *= -1
        return w

    def sample_weights(self, X, a):
        return None  # pd.Series(data=1, index=a.index)


class CleverCovariateImportanceSamplingMatrix(BaseCleverCovariate):
    """Clever covariate of inverse propensity weight vector is used as weight for
    the targeting regression. The predictors are a one-hot (full dummy) encoding
    of the treatment assignment.

    References:
        * Gruber S, van der Laan M. tmle: An R package for targeted maximum likelihood estimation. 2012.
          https://doi.org/10.18637/jss.v051.i13
    """
    def clever_covariate_fit(self, X, a):
        self.treatment_encoder_ = OneHotEncoder(sparse=False, categories="auto")
        self.treatment_encoder_.fit(a.to_frame())
        A = self.treatment_encoder_.transform(a.to_frame())
        A = pd.DataFrame(A, index=a.index, columns=self.treatment_encoder_.categories_[0])
        return A

    def clever_covariate_inference(self, X, a, treatment_value):
        treatment_assignment = np.full(
            shape=(a.shape[0], 1),
            fill_value=treatment_value,
        )
        A = self.treatment_encoder_.transform(treatment_assignment)
        A = pd.DataFrame(
            A, index=a.index, columns=self.treatment_encoder_.categories_[0]
        )
        return A

    def sample_weights(self, X, a):
        w = self.weight_model.compute_weights(X, a)
        return w


class CleverCovariateImportanceSamplingVector(BaseCleverCovariate):
    """Clever covariate of inverse propensity weight vector is used as weight for
    the targeting regression. The predictors are a signed vector with negative 1 for
    the control group.
    """
    def clever_covariate_fit(self, X, a):
        if a.nunique() != 2:
            raise AssertionError("Can only apply model on a binary treatment")
        a_sign = a.replace({0: -1})  # 2 * a - 1  # Convert a==0 to -1, keep a==1 as 1.
        return a_sign

    def clever_covariate_inference(self, X, a, treatment_value):
        treatment_value = -1 if treatment_value == 0 else treatment_value
        treatment_assignment = pd.Series(data=treatment_value, index=a.index)
        return treatment_assignment

    def sample_weights(self, X, a):
        w = self.weight_model.compute_weights(X, a)
        return w


class CleverCovariateFeatureMatrix(CleverCovariateImportanceSamplingMatrix):
    """Clever covariate uses a matrix of inverse propensity weights
    of all treatment values as a predictor to the targeting regression.

    References:
        * Gruber S, van der Laan M. tmle: An R package for targeted maximum likelihood estimation. 2012.
          https://doi.org/10.18637/jss.v051.i13
    """
    def clever_covariate_fit(self, X, a):
        A = super().clever_covariate_fit(X, a)
        W = self.weight_model.compute_weight_matrix(X, a)
        w = A * W
        return w

    def clever_covariate_inference(self, X, a, treatment_value):
        assignment = super().clever_covariate_inference(X, a, treatment_value)
        W = self.weight_model.compute_weight_matrix(X, a)
        w = assignment * W
        return w

    def sample_weights(self, X, a):
        return None  # pd.Series(data=1, index=a.index)


def _logit(p, safe=True):
    # TODO: move logit as a method, and do a clipped version with bounds specified in constructor
    if safe:
        epsilon = np.finfo(float).eps
        p = np.clip(p, epsilon, 1 - epsilon)
    return np.log(p / (1 - p))


def _clever_covariate_factory(reduced, importance_sampling) -> Type[BaseCleverCovariate]:
    if importance_sampling and reduced:
        return CleverCovariateImportanceSamplingVector
    elif importance_sampling and not reduced:
        return CleverCovariateImportanceSamplingMatrix
    elif not importance_sampling and reduced:
        return CleverCovariateFeatureVector
    else:  # not importance_sampling and not reduced
        return CleverCovariateFeatureMatrix


class TargetMinMaxScaler(MinMaxScaler):
    """A MinMaxScaler that operates on a vector (Series)"""
    # @staticmethod
    def _series_to_matrix_and_back(func):
        def to_matrix_run_and_to_series(self, X):
            X_index, X_name = X.index, X.name  # Convert back to pandas Series later
            X = X.to_frame()  # MinMaxScaler requires a 2D array, not a vector
            X = func(self, X)
            X = pd.Series(
                X.ravel(), index=X_index, name=X_name,
            )
            return X
        return to_matrix_run_and_to_series

    def fit(self, X, y=None):
        X = X.to_frame()  # MinMaxScaler requires a 2D array, not a vector
        super().fit(X, y)
        return self

    @_series_to_matrix_and_back
    def transform(self, X):
        X = super().transform(X)
        return X

    @_series_to_matrix_and_back
    def inverse_transform(self, X):
        X = super().inverse_transform(X)
        return X

    # Decorator function cannot be defined as static before decorating,
    # so setting as the decorator as `staticmethod` is done after defining the functions using the decorator
    _series_to_matrix_and_back = staticmethod(_series_to_matrix_and_back)




# (C) Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Created on Oct 24, 2019

import unittest

from pandas import DataFrame, Series

from causallib.datasets import load_nhefs, load_acic16


class BaseTestDatasets(unittest.TestCase):
    def ensure_return_types(self, loader):
        data = loader()
        self.assertTrue(isinstance(data, dict))
        self.assertTrue(hasattr(data, "X"))
        self.assertTrue(hasattr(data, "a"))
        self.assertTrue(hasattr(data, "y"))
        self.assertTrue(hasattr(data, "descriptors"))
        self.assertTrue(isinstance(data.X, DataFrame))
        self.assertTrue(isinstance(data.a, Series))
        self.assertTrue(isinstance(data.y, Series))

    def ensure_dimensions_agree(self, data):
        self.assertEqual(data.X.shape[0], data.a.shape[0])
        self.assertEqual(data.X.shape[0], data.y.shape[0])


class TestSmokingWeight(BaseTestDatasets):
    def test_return_types(self):
        self.ensure_return_types(load_nhefs)

    def test_dimensions_agree(self):
        with self.subTest("Test restrict=True"):
            data = load_nhefs(restrict=True)
            self.ensure_dimensions_agree(data)

        with self.subTest("Test restrict=False"):
            data = load_nhefs(restrict=False)
            self.ensure_dimensions_agree(data)

    def test_raw_parameter(self):
        with self.subTest("Test raw=True"):
            data = load_nhefs(raw=True)
            self.assertTrue(isinstance(data, tuple))
            self.assertEqual(len(data), 2)  # 2 = data and descriptors
            self.assertTrue(isinstance(data[0], DataFrame))
            self.assertTrue(isinstance(data[1], Series))

        with self.subTest("Test raw=False"):
            # Already asserted in test_return_Xay_parameter, return_Xay=True
            self.assertTrue(True)

    def test_restrict_parameter(self):
        with self.subTest("Test restrict=True"):
            data = load_nhefs(restrict=True)
            self.assertFalse(data.y.isnull().any())

        with self.subTest("Test restrict=False"):
            data = load_nhefs(restrict=False)
            self.assertTrue(data.y.isnull().any())


class TestACIC16(BaseTestDatasets):
    def test_return_types(self):
        self.ensure_return_types(load_acic16)
        data = load_acic16()
        self.assertTrue(hasattr(data, "po"))
        self.assertTrue(isinstance(data.po, DataFrame))

    def test_dimensions_agree(self):
        for i in range(1, 11):
            with self.subTest("Test dimension for instance {}".format(i)):
                data = load_acic16(i)
                self.ensure_dimensions_agree(data)
                self.assertEqual(data.X.shape[0], data.po.shape[0])
                self.assertEqual(data.po.shape[1], 2)

    def test_non_dummy_loading(self):
        X_dummy = load_acic16(raw=False).X
        X_factor, zymu = load_acic16(raw=True)
        self.assertEqual(X_factor.shape[0], X_dummy.shape[0])
        self.assertEqual(X_factor.shape[0], zymu.shape[0])
        self.assertEqual(5, zymu.shape[1])
        self.assertGreater(X_dummy.shape[1], X_factor.shape[1])     # Dummies has more columns



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Apr 25, 2018

"""

import abc

import pandas as pd
import numpy as np

from ..utils.general_tools import create_repr_string, get_iterable_treatment_values


class WeightEstimator:
    """
    Interface for causal estimators balancing datasets through weighting.
    """

    def __init__(self, learner, use_stabilized=False, *args, **kwargs):
        """

        Args:
            learner: Initialized sklearn model.
            use_stabilized (bool): Whether to re-weigh the learned weights with the prevalence of the treatment.
                                   See Also: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4351790/#S6title
        """
        self.learner = learner
        self.use_stabilized = use_stabilized

    @abc.abstractmethod
    def fit(self, X, a, y=None):
        """
        Trains a model to predict treatment assignment given the covariates: Pr[A|X].

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y: IGNORED.

        Returns:
            WeightEstimator: A causal weight model with an inner learner fitted.
        """
        raise NotImplementedError

    # def predict(self, X):
    #     raise NotImplementedError

    @abc.abstractmethod
    def compute_weights(self, X, a, treatment_values=None, use_stabilized=None, **kwargs):
        """
        Computes individual weight given the individual's treatment assignment.
        f(Pr[A=a_i | X_i])  for each individual i.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            treatment_values (Any | None): A desired value/s to extract weights to (i.e. weights to what treatment
                                           value should be calculated).
                                           If not specified, then the weights are chosen by the individual's actual
                                           treatment assignment.
            use_stabilized (bool): Whether to re-weigh the learned weights with the prevalence of the treatment.
                                   This overrides the use_stabilized parameter provided at initialization.
                                   See Also: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4351790/#S6title
            **kwargs:

        Returns:
            pd.Series: A vector of size (num_subjects,) with a weight for each individual
        """
        raise NotImplementedError

    @abc.abstractmethod
    def compute_weight_matrix(self, X, a, use_stabilized=None, **kwargs):
        """
        Computes individual weight across all possible treatment values.
        f(Pr[A=a_j | X_i])  for all individual i and treatment j.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            use_stabilized (bool): Whether to re-weigh the learned weights with the prevalence of the treatment.
                                   This overrides the use_stabilized parameter provided at initialization.
                                   See Also: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4351790/#S6title
            **kwargs:

        Returns:
            pd.DataFrame: A matrix of size (num_subjects, num_treatments) with weight for every individual and every
                          treatment.
        """
        raise NotImplementedError

    @staticmethod
    def _compute_stratified_weighted_aggregate(y, sample_weight=None,
                                               stratify_by=None, treatment_values=None):
        """
        Calculates aggregation of `y` weighted by `sample_weights` stratified by `stratify_by` variable.

        Args:
            y (pd.Series): The variable to aggregate (num_subjects,).
            sample_weight (pd.Series|None): Individual (sample) weights calculated.
                                            Used to achieved unbiased average outcome.
                                            If not provided, gives equal weights to every example.
            stratify_by (pd.Series|None): Categorical variable to stratify according to (num_subjects,).
                                          Namely, aggregate within subgroups sharing the same values.
                                          If not provided, the aggregation is on the entire
            treatment_values (Any): Subset of values to stratify on from `stratify_by`.
                                    If not supplied, all available stratification values are used.

        Returns:
            pd.Series[Any, float]: Series which index are treatment values, and the values are numbers - the
                                   aggregated outcome for the strata of people whose assigned treatment is the key.
        """
        if sample_weight is None:
            sample_weight = pd.Series(data=1.0, index=y.index)
        if treatment_values is None and stratify_by is None:
            stratify_by = pd.Series(data=0, index=y.index)

        treatment_values = get_iterable_treatment_values(treatment_values, stratify_by)

        res = {}
        for treatment_value in treatment_values:
            subgroup_mask = stratify_by == treatment_value
            aggregated_value = np.average(y[subgroup_mask], weights=sample_weight[subgroup_mask])
            res[treatment_value] = aggregated_value
        res = pd.Series(res)
        return res

    def evaluate_balancing(self, X, a, y, w):
        pass  # TODO: implement: (1) table one with smd (2) gather lots of metric (ks, kl, smd) (3) plot CDF of each feature

    def __repr__(self):
        repr_string = create_repr_string(self)
        return repr_string


class PropensityEstimator(WeightEstimator):
    """
    Interface for causal estimators balancing datasets through propensity (i.e. treatment probability) estimation
    (e.g. inverse probability weighting).
    """

    def __init__(self, learner, use_stabilized=False, *args, **kwargs):
        """

        Args:
            learner: Initialized sklearn model.
            use_stabilized (bool): Whether to re-weigh the learned weights with the prevalence of the treatment.
                                   See Also: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4351790/#S6title
        """
        super(PropensityEstimator, self).__init__(learner, use_stabilized=use_stabilized)
        if not hasattr(self.learner, "predict_proba"):
            raise AttributeError("Propensity Estimator must use a machine learning that can predict probabilities"
                                 "(i.e., have predict_proba method)")

    @abc.abstractmethod
    def compute_propensity(self, X, a, treatment_values=None, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def compute_propensity_matrix(self, X, a, **kwargs):
        raise NotImplementedError



import pandas as pd
import numpy as np
from typing import Optional
from sklearn.base import BaseEstimator as SKLearnBaseEstimator
from .survival_utils import safe_join
from .regression_curve_fitter import RegressionCurveFitter


class UnivariateCurveFitter:
    def __init__(self, learner: Optional[SKLearnBaseEstimator] = None):
        """
        Default implementation of a univariate survival curve fitter.
        Construct a curve fitter, either non-parametric (Kaplan-Meier) or parametric.
        API follows 'lifelines' convention for univariate models, see here for example:
        https://lifelines.readthedocs.io/en/latest/fitters/univariate/KaplanMeierFitter.html#lifelines.fitters.kaplan_meier_fitter.KaplanMeierFitter.fit
        Args:
            learner: optional scikit-learn estimator (needs to implement `predict_proba`). If provided, will
                     compute parametric curve by fitting a time-varying hazards model. if None, will compute
                     non-parametric Kaplan-Meier estimator.
        """
        self.learner = learner

    def fit(self, durations, event_observed=None, weights=None):
        """
        Fits a univariate survival curve (Kaplan-Meier or parametric, if a learner was provided in constructor)

        Args:
            durations (Iterable):  Duration subject was observed
            event_observed (Optional[Iterable]):  Boolean or 0/1 iterable, where True means 'outcome event' and False
                                                  means 'right censoring'. If unspecified, assumes that all events are
                                                  'outcome' (no censoring).
            weights (Optional[Iterable]):  Optional subject weights

        Returns:
            Self
        """
        # If 'event_observed' is unspecified, assumes that all events are 'outcome' (no censoring).
        if event_observed is None:
            event_observed = pd.Series(data=1, index=durations.index)

        if weights is None:
            weights = pd.Series(data=1, index=durations.index, name='weights')
        else:
            weights = pd.Series(data=weights, index=durations.index, name='weights')
        self.timeline_ = np.sort(np.unique(durations))

        # If sklearn classifier is provided, fit parametric curve
        if self.learner is not None:
            self.curve_fitter_ = RegressionCurveFitter(learner=self.learner)
            fit_data, (duration_col_name, event_col_name, weights_col_name) = safe_join(
                df=None, list_of_series=[durations, event_observed, weights], return_series_names=True
            )
            self.curve_fitter_.fit(df=fit_data, duration_col=duration_col_name, event_col=event_col_name,
                                   weights_col=weights_col_name)

        # Else, compute Kaplan Meier estimator non parametrically
        else:
            # Code inspired by lifelines KaplanMeierFitter
            df = pd.DataFrame({
                't': durations,
                'removed': weights.to_numpy(),
                'observed': weights.to_numpy() * (event_observed.to_numpy(dtype=bool))
            })

            death_table = df.groupby("t").sum()
            death_table['censored'] = (death_table['removed'] - death_table['observed']).astype(int)

            births = pd.DataFrame(np.zeros(durations.shape[0]), columns=["t"])
            births['entrance'] = np.asarray(weights)
            births_table = births.groupby("t").sum()
            event_table = death_table.join(births_table, how="outer", sort=True).fillna(
                0)  # http://wesmckinney.com/blog/?p=414
            event_table['at_risk'] = event_table['entrance'].cumsum() - event_table['removed'].cumsum().shift(1).fillna(
                0)
            self.event_table_ = event_table

        return self

    def predict(self, times=None, interpolate=False):
        """
        Compute survival curve for time points given in 'times' param.
        Args:
            times: sequence of time points for prediction
            interpolate: if True, linearly interpolate non-observed times. Otherwise, repeat last observed time point.

        Returns:
            pd.Series: with times index and survival values

        """
        if times is None:
            times = self.timeline_
        else:
            times = sorted(times)

        if self.learner is not None:
            # Predict parametric survival curve
            survival = self.curve_fitter_.predict_survival_function(X=None, times=pd.Series(times))
        else:
            # Compute hazard at each time step
            hazard = self.event_table_['observed'] / self.event_table_['at_risk']
            timeline = hazard.index  # if computed non-parametrically, timeline is all observed data points
            # Compute survival from hazards
            survival = pd.Series(data=np.cumprod(1 - hazard), index=timeline, name='survival')

        if interpolate:
            survival = pd.Series(data=np.interp(times, survival.index.values, survival.values),
                                 index=pd.Index(data=times, name='t'), name='survival')
        else:
            survival = survival.asof(times).squeeze()

        # Round near-zero values (may occur when using weights and all observed subjects "died" at some point)
        survival[np.abs(survival) < np.finfo(float).resolution] = 0

        return survival



"""Apply machine learning metrics to causal models for evaluation."""
import warnings

import numpy as np
import pandas as pd
from sklearn import metrics


NUMERICAL_CLASSIFICATION_METRICS = {
    "accuracy": metrics.accuracy_score,
    "precision": metrics.precision_score,
    "recall": metrics.recall_score,
    "f1": metrics.f1_score,
    "roc_auc": metrics.roc_auc_score,
    "avg_precision": metrics.average_precision_score,
    "hinge": metrics.hinge_loss,
    "matthews": metrics.matthews_corrcoef,
    "0_1": metrics.zero_one_loss,
    "brier": metrics.brier_score_loss,
}
NONNUMERICAL_CLASSIFICATION_METRICS = {
    "confusion_matrix": metrics.confusion_matrix,
    "roc_curve": metrics.roc_curve,
    "pr_curve": metrics.precision_recall_curve,
}
CLASSIFICATION_METRICS = {
    **NUMERICAL_CLASSIFICATION_METRICS,
    **NONNUMERICAL_CLASSIFICATION_METRICS,
}

REGRESSION_METRICS = {
    "expvar": metrics.explained_variance_score,
    "mae": metrics.mean_absolute_error,
    "mse": metrics.mean_squared_error,
    # "msle": metrics.mean_squared_log_error, #uncomment if predictions are all positive
    # Allow mdae receive sample_weight argument but ignore it. This unifies the interface:
    "mdae": lambda y_true, y_pred, **kwargs: metrics.median_absolute_error(
        y_true, y_pred
    ),
    "r2": metrics.r2_score,
}


def get_default_binary_metrics(only_numeric_metric=False):
    """Get default metrics for evaluating binary models.

    Args:
        only_numeric_metric (bool): If metrics_to_evaluate not provided and default is used,
            whether to use only numerical metrics. Ignored if metrics_to_evaluate is provided.
            Non-numerical metrics are for example roc_curve, that returns vectors and not scalars).
    Returns:
        dict [str, callable]: metrics dict with key: metric's name, value: callable that receives
            true labels, prediction and sample_weights (the latter is allowed to be ignored).
    """
    if only_numeric_metric:
        return NUMERICAL_CLASSIFICATION_METRICS

    return CLASSIFICATION_METRICS


def get_default_regression_metrics():
    """Get default metrics for evaluating continuous prediction models.

    Returns:
        dict [str, callable]: metrics dict with key: metric's name, value: callable that receives
            true labels, prediction and sample_weights (the latter is allowed to be ignored).
    """
    return REGRESSION_METRICS


def evaluate_metrics(
    metrics_to_evaluate,
    y_true,
    y_pred=None,
    y_pred_proba=None,
    sample_weight=None,
):
    """Evaluates the metrics against the supplied predictions and labels.

    Note that some metrics operate on proba predictions (`y_pred_proba`) and others on
    direct predictions. The function will select the correct input based on the name of the metric,
    if it knows about the metric.
    Otherwise it defaults to using the direct prediction (`y_pred`).

    Args:
        metrics_to_evaluate (dict): key: metric's name, value: callable that receives
            true labels, prediction and sample_weights (the latter is allowed to be ignored).
        y_true (pd.Series): True labels
        y_pred_proba (pd.Series): continuous output of predictor,
            as in `predict_proba` or `decision_function`.
        y_pred (pd.Series): label (i.e., categories, decisions) predictions.
        sample_weight (pd.Series | None): weight of each sample.

    Returns:
        pd.Series: name of metric as index and the evaluated score as value.
    """
    evaluated_metrics = {}
    for metric_name, metric_func in metrics_to_evaluate.items():
        prediction = y_pred_proba if _metric_needs_proba(metric_name) else y_pred
        if prediction is None:
            continue

        try:
            metric_value = metric_func(y_true, prediction, sample_weight=sample_weight)
        except ValueError as v:  # if y_true has single value
            warnings.warn(f"metric {metric_name} could not be evaluated")
            warnings.warn(str(v))
            metric_value = np.nan
        evaluated_metrics[metric_name] = metric_value

    all_scalars = all(np.isscalar(v) for v in evaluated_metrics.values())
    dtype = float if all_scalars else np.dtype(object)
    

    return pd.Series(evaluated_metrics, dtype=dtype)


def _metric_needs_proba(metric_name):
    use_proba = metric_name in {
        "hinge",
        "brier",
        "roc_curve",
        "roc_auc",
        "pr_curve",
        "avg_precision",
    }

    return use_proba



"""Evaluation results objects for plotting and further analysis.

These objects are generated by the `evaluate` method.
"""

import abc
import dataclasses
import inspect
from typing import Dict, List, Tuple, Union

import pandas as pd

from ..estimation.base_estimator import IndividualOutcomeEstimator
from ..estimation.base_weight import PropensityEstimator, WeightEstimator

from .predictions import WeightEvaluatorScores, SingleFoldPrediction
from .plots import mixins, data_extractors


@dataclasses.dataclass
class EvaluationResults(abc.ABC):
    """Data structure to hold evaluation results including cross-validation.

    Attrs:
        evaluated_metrics (Union[pd.DataFrame, WeightEvaluatorScores, None]):
        models (dict[str, Union[list[WeightEstimator], list[IndividualOutcomeEstimator]):
            Models trained during evaluation. May be dict or list or a model
            directly.
        predictions (dict[str, List[SingleFoldPredictions]]): dict with keys
            "train" and "valid" (if produced through cross-validation) and
            values of the predictions for the respective fold
        cv (list[tuple[list[int], list[int]]]): the cross validation indices,
            used to generate the results, used for constructing plots correctly
        X (pd.DataFrame): features data
        a (pd.Series): treatment assignment data
        y (pd.Series): outcome data
    """

    evaluated_metrics: Union[pd.DataFrame, WeightEvaluatorScores]
    models: Union[
        List[WeightEstimator],
        List[IndividualOutcomeEstimator],
        List[PropensityEstimator],
    ]
    predictions: Dict[str, List[SingleFoldPrediction]]
    cv: List[Tuple[List[int], List[int]]]
    X: pd.DataFrame
    a: pd.Series
    y: pd.Series

    @property
    def _extractor(self):
        """Plot-data extractor for these results.

        Implemented by child classes.
        """
        raise NotImplementedError

    @staticmethod
    def make(
        evaluated_metrics: Union[pd.DataFrame, WeightEvaluatorScores],
        models: Union[
            List[WeightEstimator],
            List[IndividualOutcomeEstimator],
            List[PropensityEstimator],
        ],
        predictions: Dict[str, List[SingleFoldPrediction]],
        cv: List[Tuple[List[int], List[int]]],
        X: pd.DataFrame,
        a: pd.Series,
        y: pd.Series,
    ):
        """Make EvaluationResults object of correct type.

        This is a factory method to dispatch the initializing data to the correct subclass of
        EvaluationResults. This is the only supported way to instantiate EvaluationResults objects.

        Args:
            evaluated_metrics (Union[pd.DataFrame, WeightEvaluatorScores]): evaluated metrics
            models (Union[
                List[WeightEstimator],
                List[IndividualOutcomeEstimator],
                List[PropensityEstimator], ]): fitted models
            predictions (Dict[str, List[SingleFoldPrediction]]): predictions by phase and fold
            cv (List[Tuple[List[int], List[int]]]): cross validation indices
            X (pd.DataFrame): features data
            a (pd.Series): treatment assignment data
            y (pd.Series): outcome data

        Raises:
            ValueError: raised if invalid estimator is passed

        Returns:
            EvaluationResults: object with results of correct type
        """

        if isinstance(models, dict):
            fitted_model = models["train"][0]
        elif isinstance(models, list):
            fitted_model = models[0]
        else:
            fitted_model = models

        if isinstance(fitted_model, PropensityEstimator):
            return PropensityEvaluationResults(
                evaluated_metrics, models, predictions, cv, X, a, y
            )
        if isinstance(fitted_model, WeightEstimator):
            return WeightEvaluationResults(
                evaluated_metrics, models, predictions, cv, X, a, y
            )
        if isinstance(fitted_model, IndividualOutcomeEstimator):
            if any(
                x and any(y.is_binary_outcome for y in x) for x in predictions.values()
            ):
                return BinaryOutcomeEvaluationResults(
                    evaluated_metrics, models, predictions, cv, X, a, y
                )
            return ContinuousOutcomeEvaluationResults(
                evaluated_metrics, models, predictions, cv, X, a, y
            )
        raise ValueError(
            f"Unable to find suitable results object for esimator of type {type(fitted_model)}"
        )

    @property
    def all_plot_names(self):
        """Available plot names.

        Returns:
            set[str]: string names of supported plot names for these results
        """
        return self._extractor.plot_names

    def get_data_for_plot(self, plot_name, phase="train"):
        """Get data for a given plot

        Args:
            plot_name (str): plot name from `self.all_plot_names`
            phase (str, optional): phase of interest. Defaults to "train".

        Returns:
            Any: the data required for the plot in question
        """
        return self._extractor.get_data_for_plot(plot_name, phase)

    def remove_spurious_cv(self):
        """Remove redundant information accumulated due to the use of cross-validation process."""
        self.models = self.models[0]
        if isinstance(self.evaluated_metrics, pd.DataFrame):
            self.evaluated_metrics.reset_index(level=["phase", "fold"], drop=True, inplace=True)
        elif isinstance(self.evaluated_metrics, WeightEvaluatorScores):
            for metric in self.evaluated_metrics:
                metric.reset_index(level=["phase", "fold"], drop=True, inplace=True)




class WeightEvaluationResults(
    EvaluationResults,
    mixins.WeightPlotterMixin,
    mixins.PlotAllMixin,
):
    __doc__ = inspect.getdoc(EvaluationResults)

    @property
    def _extractor(self):
        return data_extractors.WeightPlotDataExtractor(self)


class BinaryOutcomeEvaluationResults(
    EvaluationResults,
    mixins.ClassificationPlotterMixin,
    mixins.PlotAllMixin,
):
    __doc__ = inspect.getdoc(EvaluationResults)

    @property
    def _extractor(self):
        return data_extractors.BinaryOutcomePlotDataExtractor(self)


class ContinuousOutcomeEvaluationResults(
    EvaluationResults,
    mixins.ContinuousOutcomePlotterMixin,
    mixins.PlotAllMixin,
):
    __doc__ = inspect.getdoc(EvaluationResults)

    @property
    def _extractor(self):
        return data_extractors.ContinuousOutcomePlotDataExtractor(self)


class PropensityEvaluationResults(
    EvaluationResults,
    mixins.ClassificationPlotterMixin,
    mixins.WeightPlotterMixin,
    mixins.PlotAllMixin,
):
    __doc__ = inspect.getdoc(EvaluationResults)

    @property
    def _extractor(self):
        return data_extractors.PropensityPlotDataExtractor(self)



# (C) Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Created on Sept 25, 2019

"""
Synthetic data set generation.

This module provides data suitable for testing the HEMM estimator and for 
writing example notebooks.
"""

import numpy as np


def gen_outcomes(X, T, seed=0):
    """Synthetically generate outcome data from provided covariates and treatment assignment.

    Args:
        X: input covariates
        T: treatment assignment
        seed (int): seed to fix the random number generator
    
    Returns:
        y: observed outcomes
        mu1: true mean of treated group
        mu0: true mean of control group
        Z: indicator whether each sample belongs to positive enhanced or negative enhanced subgroup
    """
    np.random.seed(seed)
    n = X.shape[0]
    
    treated = np.where(T==1)[0]
    control = np.where(T==0)[0]

    pos = np.where(1-X[:, 1]<X[:, 0])[0]
    neg = np.where(X[:, 0]<=1-X[:, 1])[0]
    
    endist = (X[:, 0]-0.5 )**2 + (X[:, 1]-0.5 )**2
    
    posen = np.where(endist<0.1)[0]
    negen = np.where(endist>=0.1)[0]
    
    y   = np.zeros(n)
    mu1 = np.zeros(n)
    mu0 = np.zeros(n)
    
    treatposten  = list(set(treated)&set(pos)&set(posen))
    treatpostnen = list(set(treated)&set(pos)&set(negen))
    treatnegten  = list(set(treated)&set(neg)&set(posen))
    treatnegtnen = list(set(treated)&set(neg)&set(negen))
    controlpos   = list(set(control)&set(pos))
    controlneg   = list(set(control)&set(neg))
    
    y[treatposten]  = np.random.binomial(1, 0.8, len(treatposten))
    y[treatpostnen] = np.random.binomial(1, 0.6, len(treatpostnen))
    y[treatnegten]  = np.random.binomial(1, 0.6, len(treatnegten))
    y[treatnegtnen] = np.random.binomial(1, 0.4, len(treatnegtnen))
    y[controlpos]   = np.random.binomial(1, 0.4, len(controlpos))
    y[controlneg]   = np.random.binomial(1, 0.2, len(controlneg))

    mu0[pos]  = [0.4]*len(pos)
    mu0[neg]  = [0.2]*len(neg)

    posten  = list(set(pos)&set(posen))
    postnen = list(set(pos)&set(negen))
    negten  = list(set(neg)&set(posen))
    negtnen = list(set(neg)&set(negen))
    
    mu1[posten]  = [0.8]*len(posten)
    mu1[postnen] = [0.6]*len(postnen)
    mu1[negten]  = [0.6]*len(negten)
    mu1[negtnen] = [0.4]*len(negtnen)

    z = np.zeros(n)
    
    z[posen] = 1
    z[negen] = 0
    
    return y, mu1, mu0, z


def gen_treat(X, seed):
    """Generate treatment assignments for the given covariates.

    Args:
        X: input covariates
        seed: PRNG seed

    Returns:
        Individual treatment assignment.
    """
    np.random.seed(seed)

    less = np.where(X[:, 0]< 0.5)[0]
    more = np.where(X[:, 0]>=0.5)[0]
    
    treated = np.zeros(len(X))
    treated[more] = np.random.binomial(1, 0.6, len(more))
    treated[less] = np.random.binomial(1, 0.4, len(less))
    
    return treated


def gen_data(n, d, seed=0):
    """Generate outcome, treatment assignment and (d+1)-dimensional covariates for n samples.

    Args:
        n: desired number of samples.
        d: number of covariates.
        seed: PRNG seed

    Returns:
        X: generated covariates array, size = n x d+1
        T: generated treatment assignment
        Y: generated observed outcomes
        Z: generated indicator whether each sample belongs to positive enhanced or negative enhanced subgroup
        mu1: true mean counterfactual outcome under treatment
        mu0: true mean counterfactual outcome under control
    """
    np.random.seed(seed)
    
    treatment_group_size = n//2
    Xt = np.random.rand(treatment_group_size, d)
    Xc = np.random.rand(treatment_group_size, d)
    X = np.vstack([Xt, Xc])
    ones = np.array([np.ones(len(X))]).T
    X = np.hstack([X, ones])
    
    T = gen_treat(X, seed)
    
    Y, mu1, mu0, Z = gen_outcomes(X, T, seed)
    
    return X, T, Y, Z, mu1, mu0


def gen_montecarlo(n, d, num_repeats=100):
    """Generate synthetic train and test data (80:20 proportions).

    Args:
        n: desired number of samples.
        d: dimensionality of each samples (number of covariates) - 1
        num_repeats: number of subgroups to generate, each repeat will produce n covariates with dimension d + 1.

    Returns:
        Dictionary holding test and training data. Each data slot has the form:
            x: generated covariates array, size = n x d+1 x num_repeats
            t: generated treatment assignment for covariates, size = n x num_repeats
            yf: generated observed outcomes for each covariate, size = n x num_repeats
            Z: generated indicator whether each sample belongs to positive enhanced or negative enhanced subgroup.
               size = n x num_repeats
            mu1: true mean counterfactual outcome under treatment, size = n x num_repeats
            mu0: true mean counterfactual outcome under control, size = n x num_repeats
    """
    Xtr, Ttr, Ytr, Ztr, mu1tr, mu0tr = [], [], [], [], [], []
    Xte, Tte, Yte, Zte, mu1te, mu0te = [], [], [], [], [], []

    for repeat in range(num_repeats):
        X_, T_, Y_, Z_, mu1_, mu0_ = gen_data(n, d, repeat)
        
        # Of the n data points to generate, use 80% for training and 20% for testing.
        r = (4*n)//5
        
        X_tr, T_tr, Y_tr, Z_tr, mu1_tr, mu0_tr = X_[:r], T_[:r], Y_[:r], Z_[:r], mu1_[:r], mu0_[:r]
        X_te, T_te, Y_te, Z_te, mu1_te, mu0_te = X_[r:], T_[r:], Y_[r:], Z_[r:], mu1_[r:], mu0_[r:]

        Xtr.append(X_tr.T)
        Ttr.append(T_tr)
        Ytr.append(Y_tr)
        Ztr.append(Z_tr)
        mu1tr.append(mu1_tr)
        mu0tr.append(mu0_tr)
        
        Xte.append(X_te.T)
        Tte.append(T_te)
        Yte.append(Y_te)
        Zte.append(Z_te)
        mu1te.append(mu1_te)
        mu0te.append(mu0_te)   
    
    data = {}
    data['TRAIN'] = {}
    data['TEST'] = {}

    data['TRAIN']['x'] = np.array(Xtr).T
    data['TRAIN']['t'] = np.array(Ttr).T
    data['TRAIN']['yf'] = np.array(Ytr).T
    data['TRAIN']['z'] = np.array(Ztr).T
    data['TRAIN']['mu1'] = np.array(mu1tr).T
    data['TRAIN']['mu0'] = np.array(mu0tr).T
    
    data['TEST']['x'] = np.array(Xte).T
    data['TEST']['t'] = np.array(Tte).T
    data['TEST']['yf'] = np.array(Yte).T
    data['TEST']['z'] = np.array(Zte).T
    data['TEST']['mu1'] = np.array(mu1te).T
    data['TEST']['mu0'] = np.array(mu0te).T
    
    return data



from setuptools import setup, find_packages
import os


GIT_URL = "https://github.com/IBM/causallib"

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


def get_lines(filename):
    with open(filename, 'r') as f:
        return f.read().splitlines()


def get_valid_packages_from_requirement_file(file_path):
    lines = get_lines(file_path)
    # Filter out non-package lines that are legal for `pip install -r` but fail for setuptools' `require`:
    pkg_list = [p for p in lines if p.lstrip()[0].isalnum()]
    return pkg_list


def get_version(filename):
    with open(filename, 'r') as f:
        for line in f.readlines():
            if line.startswith('__version__'):
                quotes_type = '"' if '"' in line else "'"
                version = line.split(quotes_type)[1]
                return version
    raise RuntimeError("Unable to find version string.")


setup(name='causallib',
      version=get_version(os.path.join('causallib', '__init__.py')),
      # packages=find_packages(exclude=['scripts', 'data', 'tests']),
      packages=find_packages(),
      description='A library of causal inference tools by IBM Haifa Research Labs',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url=GIT_URL,
      author='IBM Research Haifa Labs - Machine Learning for Healthcare and Life Sciences',
      # author_email=None,
      license="Apache License 2.0",
      keywords="causal inference effect estimation causality",
      install_requires=get_valid_packages_from_requirement_file("requirements.txt"),
      extras_require={
          'contrib': get_valid_packages_from_requirement_file(os.path.join("causallib", "contrib", "requirements.txt")),
          'docs': get_valid_packages_from_requirement_file(os.path.join("docs", "requirements.txt"))
      },
      # include_package_data=True,
      package_data={
          'causallib': [os.path.join('datasets', 'data', '*/*.csv')]
      },
      project_urls={
          'Documentation': 'https://causallib.readthedocs.io/en/latest/',
          'Source Code': GIT_URL,
          'Bug Tracker': GIT_URL + '/issues',
      },
      classifiers=[
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
          "License :: OSI Approved :: Apache Software License",
          "Development Status :: 4 - Beta",
          "Topic :: Scientific/Engineering",
          "Intended Audience :: Science/Research"
      ]
      )



from .survival_utils import canonize_dtypes_and_names
from .standardized_survival import StandardizedSurvival
from causallib.estimation.base_weight import WeightEstimator
import pandas as pd
from typing import Any, Optional


class WeightedStandardizedSurvival(StandardizedSurvival):
    """
    Combines WeightedSurvival and StandardizedSurvival:
        1. Adjusts for treatment assignment by creating weighted pseudo-population (e.g., inverse propensity weighting).
        2. Computes parametric curve by fitting a time-varying hazards model that includes baseline covariates.
    """

    def __init__(self,
                 weight_model: WeightEstimator,
                 survival_model: Any,
                 stratify: bool = True):
        """
        Combines WeightedSurvival and StandardizedSurvival:
            1. Adjusts for treatment assignment by creating weighted pseudo-population (e.g., inverse propensity weighting).
            2. Computes parametric curve by fitting a time-varying hazards model that includes baseline covariates.
        Args:
            weight_model: causallib compatible weight model (e.g., IPW)
           survival_model: Two alternatives:
                1. Scikit-Learn estimator (needs to implement `predict_proba`) - compute parametric curve by fitting a
                    time-varying hazards model that includes baseline covariates. Note that the model is fitted on a
                    person-time table with all covariates, and might be computationally and memory expansive.
                2. lifelines RegressionFitter - use lifelines fitter to compute survival curves from baseline covariates,
                    events and durations
            stratify (bool): if True, fit a separate model per treatment group
        """
        self.weight_model = weight_model
        super().__init__(survival_model=survival_model, stratify=stratify)

    def fit(self,
            X: pd.DataFrame,
            a: pd.Series,
            t: pd.Series,
            y: pd.Series,
            w: Optional[pd.Series] = None,
            fit_kwargs: Optional[dict] = None):
        """
        Fits parametric models and calculates internal survival functions.

        Args:
            X (pd.DataFrame): Baseline covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            t (pd.Series): Followup duration, size (num_subjects,).
            y (pd.Series): Observed outcome (1) or right censoring event (0), size (num_subjects,).
            w (pd.Series): NOT USED (for compatibility only) optional subject weights.
            fit_kwargs (dict): Optional kwargs for fit call of survival model

        Returns:
            self
        """
        a, t, y, _, X = canonize_dtypes_and_names(a=a, t=t, y=y, w=None, X=X)
        self.weight_model.fit(X=X, a=a, y=y)
        iptw_weights = self.weight_model.compute_weights(X, a)

        # Call fit from StandardizedSurvival, with added ipt weights
        super().fit(X=X, a=a, t=t, y=y, w=iptw_weights, fit_kwargs=fit_kwargs)
        return self





from .weighted_survival import WeightedSurvival
from typing import Any


class MarginalSurvival(WeightedSurvival):
    """
    Marginal (un-adjusted) survival estimator.
    Essentially it is a degenerated WeightedSurvival instance without a weight model.
    """
    def __init__(self,
                 survival_model: Any = None):
        """
        Marginal (un-adjusted) survival estimator.
        Args:
            survival_model: Three alternatives:
                    1. None - compute non-parametric KaplanMeier survival curve
                    2. Scikit-Learn estimator (needs to implement `predict_proba`) - compute parametric curve by fitting a time-varying hazards model
                    3. lifelines UnivariateFitter - use lifelines fitter to compute survival curves from events and durations
        """

        super().__init__(weight_model=None, survival_model=survival_model)




"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Jun 09, 2021

"""

import warnings

import pandas as pd

from causallib.estimation.base_estimator import PopulationOutcomeEstimator
from causallib.estimation.base_weight import PropensityEstimator
from causallib.estimation.ipw import IPW

# TODO: Move fit and _predict methods to PropensityEstimator, instead of rely on it from the ipw module.


class OverlapWeights(IPW):

    def __init__(self, learner, use_stabilized=False):
        """
        Implementation of overlap (propensity score) weighting:

        https://www.tandfonline.com/doi/full/10.1080/01621459.2016.1260466

        A method to balance observed covariates between treatment groups in observational studies.
        Down-weigh observations with extreme propensity and weigh up
        Put less importance to observations with extreme propensity scores,
        and put more emphasis on observations with a central tendency towards
        (i.e. overlapping propensity scores).

        Each unit’s weight is proportional to the probability of that unit being
        assigned to the opposite group:
        w_i = 1 - Pr[A=a_i|Xi]

        This method assumes only two treatment groups exist.

        Args:
            learner: Initialized sklearn model.
            use_stabilized (bool): Whether to re-weigh the learned weights with the prevalence of the treatment.
                                   See Also: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4351790/#S6title
        """
        super(OverlapWeights, self).__init__(learner, use_stabilized)

    def compute_weight_matrix(self, X, a, clip_min=None, clip_max=None, use_stabilized=None):
        """
        Computes individual weight across all possible treatment values.
        w_ij = 1 - Pr[A=a_j | X_i]  for all individual i and treatment j.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            clip_min (None|float): Lower bound for propensity scores. Better be left `None`.
            clip_max (None|float): Upper bound for propensity scores. Better be left `None`.
            use_stabilized (None|bool): Whether to re-weigh the learned weights with the prevalence of the treatment.
                                        This overrides the use_stabilized parameter provided at initialization.
                                        If True provided, but the model was initialized with use_stabilized=False, then
                                        prevalence is calculated from data at hand, rather than the prevalence from the
                                        training data.
                                        See Also: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4351790/#S6title

        Returns:
            pd.DataFrame: A matrix of size (num_subjects, num_treatments) with weight for every individual and every
                          treatment.
        """
        use_stabilized = self.use_stabilized if use_stabilized is None else use_stabilized
        # Check that number of unique classes is 2
        self.__check_number_of_classes_is_legal(a)
        # Truncation is generally bad, check and warn:
        self.__check_truncation_value_is_none(clip_min, clip_max)

        # COmpute propensity scores
        probabilities = self.compute_propensity_matrix(X, a, clip_min, clip_max)
        # weight matrix: 1-P[a_i=1|x]
        # Reverse probabilities to opposite classes:
        probabilities.columns = probabilities.columns[::-1]  # Flip name-based indexing
        # reorder weights_matrix
        weight_matrix = probabilities.iloc[:, ::-1]  # Flip integer (location)-based indexing
        weight_matrix = self.stabilize_weights(a, weight_matrix, use_stabilized)

        return weight_matrix

    def stabilize_weights(self, a, weight_matrix, use_stabilized=False):
        # TODO: Move this function to IPW
        """
             Adjust sample weights according to class prevalence:
             Pr[A=a_i] * w_i

             Args:
                 weight_matrix (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
                 use_stabilized (None|bool): Whether to re-weigh the learned weights with the prevalence of the treatment.
                                             This overrides the use_stabilized parameter provided at initialization.
                                             If True provided, but the model was initialized with use_stabilized=False, then
                                             prevalence is calculated from data at hand, rather than the prevalence from the
                                             training data.
                                             See Also: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4351790/#S6title

             Returns:
                 pd.DataFrame: A matrix of size (num_subjects, num_treatments) with stabilized (if True)
                                weight for every individual and every treatment.
             """
        if use_stabilized:
            if self.use_stabilized:
                prevalence = self.treatment_prevalence_
            else:
                warnings.warn("Stabilized is asked, however, the model was not trained using stabilization, and "
                              "therefore, stabilized weights are taken from the provided treatment assignment.",
                              RuntimeWarning)
                prevalence = a.value_counts(normalize=True, sort=False)
            prevalence_per_subject = a.replace(prevalence)  # map tx-assign to prevalence
            # pointwise multiplication of each column in weights:
            weight_matrix = weight_matrix.multiply(prevalence_per_subject, axis="index")

        return weight_matrix

    @staticmethod
    def __check_number_of_classes_is_legal(x):
        count_classes = x.nunique()
        if count_classes != 2:
            raise AssertionError("Number of unique classes should be equal 2")

    @staticmethod
    def __check_truncation_value_is_none(clip_min, clip_max):
        if clip_min is not None or clip_max is not None:
            warnings.warn(
                "Trimming observations with Overlap Weighting may be redundant, "
                "as extreme observations can receive greater importance than they should.",
                RuntimeWarning
            )



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Apr 25, 2018

"""

import warnings

import pandas as pd

from .base_estimator import PopulationOutcomeEstimator
from .base_weight import PropensityEstimator
from ..utils.stat_utils import robust_lookup


class IPW(PropensityEstimator, PopulationOutcomeEstimator):
    """
    Causal model implementing inverse probability (propensity score) weighting.
    w_i = 1 / Pr[A=a_i|Xi]
    """

    def __init__(self, learner, clip_min=None, clip_max=None, use_stabilized=False, verbose=False):
        """

        Args:
            learner: Initialized sklearn model.
            clip_min (None|float): Optional value between 0 to 0.5 to lower bound the propensity estimation by
                                   clipping it. Will clip probabilities under clip_min to this value.
            clip_max (None|float): Optional value between 0.5 to 1 to upper bound the propensity estimation by
                                   clipping it. Will clip probabilities above clip_max to this value.
            use_stabilized (bool): Whether to re-weigh the learned weights with the prevalence of the treatment.
                                   See Also: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4351790/#S6title
            verbose (bool): Whether to print summary statistics regarding the number of samples clipped due to
                            clip_min and clip_max.
        """
        super(IPW, self).__init__(learner, use_stabilized)
        self.__check_truncation_value_is_legal(clip_min, clip_max)
        self.clip_min = clip_min
        self.clip_max = clip_max
        self.verbose = verbose

    def fit(self, X, a, y=None):
        if self.use_stabilized:
            self.treatment_prevalence_ = a.value_counts(normalize=True, sort=False)
        self.learner.fit(X, a)
        return self

    def _predict(self, X):
        # Assumes PropensityEstimator checked that learner has predict_proba during initialization:
        prediction_matrix = self.learner.predict_proba(X)
        columns = getattr(self.learner, "classes_", None)  # non-sklearn classifiers might not have the attribute
        prediction_matrix = pd.DataFrame(prediction_matrix, index=X.index, columns=columns)
        return prediction_matrix

    def compute_weights(self, X, a, treatment_values=None, clip_min=None, clip_max=None, use_stabilized=None):
        """
        Computes individual weight given the individual's treatment assignment.
        w_i = 1 / Pr[A=a_i|X_i]  for each individual i.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            treatment_values (Any | None): A desired value/s to extract weights to (i.e. weights to what treatment
                                           value should be calculated).
                                           If not specified, then the weights are chosen by the individual's actual
                                           treatment assignment.
            clip_min (None|float): Optional value between 0 to 0.5 to lower bound the propensity estimation by
                                   clipping it. Will clip probabilities under clip_min to this value.
            clip_max (None|float): Optional value between 0.5 to 1 to upper bound the propensity estimation by
                                   clipping it. Will clip probabilities above clip_max to this value.
            use_stabilized (None|bool): Whether to re-weigh the learned weights with the prevalence of the treatment.
                                        This overrides the use_stabilized parameter provided at initialization.
                                        If True provided, but the model was initialized with use_stabilized=False, then
                                        prevalence is calculated from data at hand, rather than the prevalence from the
                                        training data.
                                        See Also: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4351790/#S6title

        Returns:
            pd.Series | pd.DataFrame: If treatment_values is not supplied (None) or is a scalar, then a vector of
                                      n_samples with a weight for each sample is returned.
                                      If treatment_values is a list/array, then a DataFrame is returned.
        """
        weight_matrix = self.compute_weight_matrix(X, a, clip_min, clip_max, use_stabilized)
        if treatment_values is None:
            weights = robust_lookup(weight_matrix, a)  # lookup table: take the column a[i] for every i in index(a).
        else:
            weights = weight_matrix[treatment_values]
        return weights

    def compute_weight_matrix(self, X, a, clip_min=None, clip_max=None, use_stabilized=None):
        """
        Computes individual weight across all possible treatment values.
        w_ij = 1 / Pr[A=a_j | X_i]  for all individual i and treatment j.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            clip_min (None|float): Optional value between 0 to 0.5 to lower bound the propensity estimation by clipping it.
                                       Will clip probabilities under clip_min to this value.
            clip_max (None|float): Optional value between 0.5 to 1 to upper bound the propensity estimation by clipping it.
                                       Will clip probabilities above clip_max to this value.
            use_stabilized (None|bool): Whether to re-weigh the learned weights with the prevalence of the treatment.
                                        This overrides the use_stabilized parameter provided at initialization.
                                        If True provided, but the model was initialized with use_stabilized=False, then
                                        prevalence is calculated from data at hand, rather than the prevalence from the
                                        training data.
                                        See Also: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4351790/#S6title

        Returns:
            pd.DataFrame: A matrix of size (num_subjects, num_treatments) with weight for every individual and every
                          treatment.
        """
        use_stabilized = self.use_stabilized if use_stabilized is None else use_stabilized

        probabilities = self.compute_propensity_matrix(X, a, clip_min, clip_max)

        # weight_matrix = 1.0 / probabilities                                                     # type: pd.DataFrame
        weight_matrix = probabilities.rdiv(1.0)

        if use_stabilized:
            if self.use_stabilized:
                prevalence = self.treatment_prevalence_
            else:
                warnings.warn("Stabilized is asked, however, the model was not trained using stabilization, and "
                              "therefore, stabilized weights are taken from the provided treatment assignment.",
                              RuntimeWarning)
                prevalence = a.value_counts(normalize=True, sort=False)
            prevalence_per_subject = a.replace(prevalence)  # map tx-assign to prevalence
            # pointwise multiplication of each column in weights:
            weight_matrix = weight_matrix.multiply(prevalence_per_subject, axis="index")

        return weight_matrix

    def compute_propensity(self, X, a, treatment_values=None, clip_min=None, clip_max=None):
        """

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            treatment_values (Any | None): A desired value/s to extract propensity to (i.e. probabilities to what
                                          treatment value should be calculated).
                                          If not specified, then the maximal treatment value is chosen. This is since
                                          the usual case is of treatment (A=1) control (A=0) setting.
            clip_min (None|float): Optional value between 0 to 0.5 to lower bound the propensity estimation by clipping it.
                                       Will clip probabilities under clip_min to this value.
            clip_max (None|float): Optional value between 0.5 to 1 to upper bound the propensity estimation by clipping it.
                                       Will clip probabilities above clip_max to this value.

        Returns:
            pd.DataFrame | pd.Series: A matrix/vector num_subjects rows and number of columns is the number of values
                                      provided to treatment_value. The content is probabilities for every individual
                                      to have the specified treatment_value.
                                      If treatment_value is a list/vector, than a pd.DataFrame is returned.
                                      If treatment_value is sort of scalar, than a pd.Series is returned.
                                       (just like slicing a DataFrame's columns)
        """
        treatment_values = a.max() if treatment_values is None else treatment_values

        probabilities = self.compute_propensity_matrix(X, a, clip_min, clip_max)
        probabilities = probabilities[treatment_values]
        return probabilities

    def compute_propensity_matrix(self, X, a=None, clip_min=None, clip_max=None):
        """

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            clip_min (None|float): Optional value between 0 to 0.5 to lower bound the propensity estimation by clipping it.
                                       Will clip probabilities under clip_min to this value.
            clip_max (None|float): Optional value between 0.5 to 1 to upper bound the propensity estimation by clipping it.
                                       Will clip probabilities above clip_max to this value.

        Returns:
            pd.DataFrame: A matrix of size (num_subjects, num_treatments) with probability for every individual and e
                          very treatment.
        """
        clip_min = self.clip_min if clip_min is None else clip_min
        clip_max = self.clip_max if clip_max is None else clip_max
        self.__check_truncation_value_is_legal(clip_min, clip_max)

        probabilities = self._predict(X)

        clip_min = 0 if clip_min is None else clip_min
        clip_max = 1 if clip_max is None else clip_max
        self.__count_truncated(probabilities, clip_min, clip_max)
        probabilities = probabilities.clip(lower=clip_min, upper=clip_max)

        return probabilities

    def estimate_population_outcome(self, X, a, y, w=None, treatment_values=None):
        """
        Calculates weighted population outcome for each subgroup stratified by treatment assignment.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y (pd.Series): Observed outcome of size (num_subjects,).
            w (pd.Series | None): Individual (sample) weights calculated. Used to achieved unbiased average outcome.
                                   If not provided, will be calculated on the data.
            treatment_values (Any): Desired treatment value/s to stratify upon.
                                    Must be a subset of values found in `a`.
                                    If not supplied, calculates for all available treatment values.

        Returns:
            pd.Series[Any, float]: Series which index are treatment values, and the values are numbers - the
                                   aggregated outcome for the strata of people whose assigned treatment is the key.
        """
        if w is None:
            w = self.compute_weights(X, a)
        res = self._compute_stratified_weighted_aggregate(y, sample_weight=w, stratify_by=a,
                                                          treatment_values=treatment_values)
        return res

    @staticmethod
    def __check_truncation_value_is_legal(clip_min, clip_max):
        if clip_min is not None and not 0 <= clip_min <= 0.5:
            raise AssertionError("Provided value for truncation (clip_min) should be between 0.0 and 0.5")
        if clip_max is not None and not 0.5 <= clip_max <= 1:
            raise AssertionError("Provided value for truncation (clip_max) should be between 0.5 and 1.0")

    def __count_truncated(self, probabilities, clip_min, clip_max):
        num_clipped = probabilities.apply(lambda x: ~x.between(clip_min, clip_max)).sum().sum()
        fraction_clipped = num_clipped/probabilities.size
        if self.verbose:
            print(f'Fraction of values being truncated: {fraction_clipped:.5f}.')  # TODO: do as log
        return fraction_clipped



from .doubly_robust import AIPW, PropensityFeatureStandardization, WeightedStandardization
from .ipw import IPW
from .overlap_weights import OverlapWeights
from .standardization import Standardization, StratifiedStandardization
from .marginal_outcome import MarginalOutcomeEstimator
from .matching import Matching, PropensityMatching
from .rlearner import RLearner
from .xlearner import XLearner
from .tmle import TMLE




# (C) Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Created on Oct 24, 2019

from causallib.utils.stat_utils import robust_lookup
import os
import pandas as pd
from sklearn.utils import Bunch

DATA_DIR_NAME = "data"


def load_data_file(file_name, data_dir_name, sep=","):
    module_path = os.path.dirname(__file__)
    file_path = os.path.join(module_path, data_dir_name, file_name)
    data = pd.read_csv(file_path, sep=sep)
    return data


def load_nhefs(raw=False, restrict=True, augment=True, onehot=True):
    """Loads the NHEFS smoking-cessation and weight-loss dataset.

    Data was gathered during an observational study conducted by the NHANS
    during the 1970's and 1980'. It follows a cohort a people whom some
    decided to quite smoking and some decided to persist, and record the
    gain in weight for each individual to try estimate the causal contribution
    of smoking cessation on weight gain.

    This dataset is used throughout Hernán and Robins' Causal Inference Book.
     https://www.hsph.harvard.edu/miguel-hernan/causal-inference-book/
    If used for academic purposes, please consider citing the book:
     Hernán MA, Robins JM (2020). Causal Inference: What If. Boca Raton: Chapman & Hall/CRC.

    Args:
        raw (bool): Whether to return the entire DataFrame and descriptors or not.
                    If False, only confounders are used for the data.
                    If True, returns a (pd.DataFrame, pd.Series) tuple (data and description).
        restrict (bool): Whether to apply exclusion criteria on missing data or not.
                         Note: if False - data will have censored (NaN) outcomes.
        augment (bool): Whether to add augmented (squared) features
                    If False, only original data returned.
                    If True, squares continuous valued columns ['age', 'wt71', 'smokeintensity', 'smokeyrs']
                    and joins to data frame with suffix '^2'
        onehot (bool): Whether to one-hot encode categorical data.
                    If False, categorical data ["active", "education", "exercise"], will be returned
                    in individual columns with categorical values.
                    If True, extra columns with the categorical value one-hot encoded.

    Returns:
        Bunch: dictionary-like object
               attributes are: `X` (covariates), `a` (treatment assignment) `y` (outcome),
                               `descriptors` (feature description)
    """
    dir_name = os.path.join(DATA_DIR_NAME, "nhefs")
    data = load_data_file("NHEFS.csv", dir_name)
    descriptors = load_data_file("NHEFS_codebook.csv", dir_name)

    descriptors = descriptors.set_index("Variable name")["Description"]

    if raw:
        return data, descriptors

    confounders = ["active", "age", "education", "exercise", "race",
                   "sex", "smokeintensity", "smokeyrs", "wt71"]

    if restrict:
        restrictions = ["wt82"]
        missing = data[restrictions].isnull().any(axis="columns")
        data = data.loc[~missing]

    a = data.pop("qsmk")
    y = data.pop("wt82_71")
    X = data[confounders]
    descriptors = descriptors[confounders + ["qsmk", "wt82_71"]]
    if onehot:
        X = pd.get_dummies(
            X, columns=["active", "education", "exercise"], drop_first=True)
    if augment:
        X = X.join(X[['age', 'wt71', 'smokeintensity', 'smokeyrs']]
                   ** 2, rsuffix="^2")

    data = Bunch(X=X, a=a, y=y, descriptors=descriptors)
    return data


def load_nhefs_survival(augment=True, onehot=True):
    """
    Loads and pre-processes the NHEFS smoking-cessation dataset.

    Data was gathered in an observational study conducted by the NHANS
    during the 1970's and 1980'. It follows a cohort a people whom some
    decided to quite smoking and some decided to persist, and record the
    death events within 10 years of follow-up.

    This dataset is used throughout Hernán and Robins' Causal Inference Book.
    https://www.hsph.harvard.edu/miguel-hernan/causal-inference-book/
    If used for academic purposes, please consider citing the book:
    Hernán MA, Robins JM (2020). Causal Inference: What If. Boca Raton: Chapman & Hall/CRC.

    Args:
        augment (bool): Whether to add augmented (squared) features
                    If False, only original data returned.
                    If True, squares continuous valued columns ['age', 'wt71', 'smokeintensity', 'smokeyrs']
                    and joins to data frame with suffix '^2'
        onehot (bool): Whether to one-hot encode categorical data.
                    If False, categorical data ["active", "education", "exercise"], will be returned
                    in individual columns with categorical values.
                    If True, extra columns with the categorical value one-hot encoded.

    Returns:
        X (pd.DataFrame): Baseline covariate matrix of size (num_subjects, num_features).
        a (pd.Series): Treatment assignment of size (num_subjects,). Quit smoking vs. non-quit.
        t (pd.Series): Followup duration, size (num_subjects,).
        y (pd.Series): Observed outcome (1) or right censoring event (0), size (num_subjects,).
    """

    nhefs_all = load_nhefs(raw=True, augment=augment, onehot=onehot)[0]

    nhefs_all['longevity'] = (nhefs_all.yrdth - 83) * 12 + nhefs_all.modth - 1
    nhefs_all['longevity'].fillna(120, inplace=True)

    # Pre-process data
    a = nhefs_all['qsmk']
    t = nhefs_all['longevity']
    y = nhefs_all['death']
    X = nhefs_all[[
        "sex", "race", "age",
        "active", "education", "exercise",
        "smokeintensity", "smokeyrs",
        "wt71"
    ]]

    # Add square terms and dummy variables
    squares = {}
    for col in ['age', 'wt71', 'smokeintensity', 'smokeyrs']:
        squares[f'{col}^2'] = X[col] * X[col]
    X = X.assign(**squares)
    X = pd.get_dummies(
        X, columns=["active", "education", "exercise"], drop_first=True
    )

    # Make timeline 1-index (to comply with some lifelines fitters that require strictly positive time steps)
    t = t + 1

    data = Bunch(X=X, a=a, t=t, y=y)
    return data


def load_acic16(instance=1, raw=False):
    """ Loads single dataset from the 2016 Atlantic Causal Inference Conference data challenge.

    The dataset is based on real covariates but synthetically simulates the treatment assignment
    and potential outcomes. It therefore also contains sufficient ground truth to evaluate
    the effect estimation of causal models.
    The competition introduced 7700 simulated files (100 instances for each of the 77
    data-generating-processes). We provide a smaller sample of one instance from 10
    DGPs. For the full dataset, see the link below to the competition site.

    If used for academic purposes, please consider citing the competition organizers:
     Vincent Dorie, Jennifer Hill, Uri Shalit, Marc Scott, and Dan Cervone. "Automated versus do-it-yourself methods
     for causal inference: Lessons learned from a data analysis competition."
     Statistical Science 34, no. 1 (2019): 43-68.

    Args:
        instance (int): number between 1-10 (inclusive), dataset to load.
        raw (bool): Whether to apply contrast ("dummify") on non-numeric columns
                    If True, returns a (pd.DataFrame, pd.DataFrame) tuple (one for covariates and the second with
                    treatment assignment, noisy potential outcomes and true potential outcomes).

    Returns:
        Bunch: dictionary-like object
               attributes are: `X` (covariates), `a` (treatment assignment), `y` (outcome),
                               `po` (ground truth potential outcomes: `po[0]` potential outcome for controls and
                                `po[1]` potential outcome for treated),
                               `descriptors` (feature description).


    See Also:
        * `Publication <https://projecteuclid.org/euclid.ss/1555056030>`_
        * `Official competition site <http://jenniferhill7.wixsite.com/acic-2016/competition>`_
        * `Official github with data generating code <https://github.com/vdorie/aciccomp/tree/master/2016>`_
    """
    dir_name = os.path.join(DATA_DIR_NAME, "acic_challenge_2016")

    X = load_data_file("x.csv", dir_name)
    zymu = load_data_file("zymu_{}.csv".format(instance), dir_name)

    if raw:
        return X, zymu

    non_numeric_cols = X.select_dtypes(include=[object]).columns
    X = pd.get_dummies(X, columns=non_numeric_cols, drop_first=True)

    a = zymu["z"].rename("a")
    # # Extract observed outcome:
    y = zymu[["y0", "y1"]]
    y = y.rename(columns=lambda x: int(x.strip("y")))  # remove 'y' prefix to allow lookup
    y = robust_lookup(y, a)
    # # Potential outcomes:
    po = zymu[["mu0", "mu1"]]
    po = po.rename(columns=lambda x: x.strip("mu"))

    descriptors = pd.Series(data="No true meaning", index=X.columns)
    data = Bunch(X=X, a=a, y=y, po=po, descriptors=descriptors)
    return data



import pandas as pd
from abc import ABC, abstractmethod


class SurvivalBase(ABC):
    """
    Interface class for causal survival analysis with fixed baseline covariates.
    """
    @abstractmethod
    def fit(self,
            X: pd.DataFrame,
            a: pd.Series,
            t: pd.Series,
            y: pd.Series):
        """
        Fits internal learner(s).

        Args:
            X (pd.DataFrame): Baseline covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            t (pd.Series): Followup duration, size (num_subjects,).
            y (pd.Series): Observed outcome (1) or right censoring event (0), size (num_subjects,).

        Returns:
            self
        """
        raise NotImplementedError

    @abstractmethod
    def estimate_population_outcome(self,
                                    **kwargs) -> pd.DataFrame:
        """
        Returns population averaged survival curves.

        Returns:
            pd.DataFrame: with time-step index, treatment values as columns and survival as entries
        """
        raise NotImplementedError


class SurvivalTimeVaryingBase(SurvivalBase):
    """
    Interface class for causal survival analysis estimators that support time-varying followup covariates.
    Followup covariates matrix (XF) needs to have a 'time' column, and indexed by subject IDs that correspond to
    the other inputs (X, a, y, t). All columns other than 'time' will be used for time-varying adjustments.

    Example XF format:
    +----+------+------+------+
    | id |   t  | var1 | var2 |
    +----+------+------+------+
    |  1 |    0 |  1.4 |   22 |
    |  1 |    4 |  1.2 |   22 |
    |  1 |    8 |  1.5 |  NaN |
    |  2 |    0 |  1.6 |   10 |
    |  2 |   11 |  1.6 |   11 |
    +----+------+------+------+
    """

    @abstractmethod
    def fit(self,
            X: pd.DataFrame,
            a: pd.Series,
            t: pd.Series,
            y: pd.Series,
            XF: pd.DataFrame = None) -> None:
        """
        Fits internal survival functions.

        Args:
            X (pd.DataFrame): Baseline covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            t (pd.Series): Followup duration, size (num_subjects,).
            y (pd.Series): Observed outcome (1) or right censoring event (0), size (num_subjects,).
            XF (pd.DataFrame): Time-varying followup covariate matrix

        Returns:
            A fitted estimator with precalculated survival functions.
        """
